"""
Py-ART: The Python ARM Radar Toolkit
=====================================

"""

# Detect if we're being called as part of Py-ART's setup procedure
try:
    __PYART_SETUP__
except NameError:
    __PYART_SETUP__ = False

if __PYART_SETUP__:
    import sys as _sys
    _sys.stderr.write("Running from Py-ART source directory.\n")
    del _sys
else:

    # print out helpful message if build fails or importing from source tree
    from . import __check_build

    # versioning
    from .version import git_revision as __git_revision__
    from .version import version as __version__

    # import subpackages
    from . import core
    from . import io
    from . import correct
    from . import graph
    from . import map
    from . import testing
    from . import config
    from . import aux_io
    from . import retrieve

    # root level functions
    from .config import load_config
    from ._debug_info import _debug_info

    # test function setup based on scikit-image test function
    import imp as _imp
    import os.path as _osp
    import functools as _functools

    try:
        _imp.find_module('nose')
    except ImportError:
        def _test(verbose=False):
            """
            This would invoke the Py-ART test suite, but nose couldn't be
            imported so the test suite can not run.
            """
            raise ImportError("Could not load nose. Unit tests not available.")
    else:
        def _test(verbose=False):
            """
            Invoke the Py-ART test suite.
            """
            import nose
            pkg_dir = _osp.abspath(_osp.dirname(__file__))
            args = ['', pkg_dir, '--exe']
            if verbose:
                args.extend(['-v', '-s'])
            nose.run('pyart', argv=args)

    # do not use `test` as function name as this leads to a recursion problem
    # with the nose test suite
    test = _test
    test_verbose = _functools.partial(test, verbose=True)
    test_verbose.__doc__ = test.__doc__



"""
======================================================
Create a multiple panel RHI plot from a CF/Radial file
======================================================

An example which creates a multiple panel RHI plot of a CF/Radial file using
both a CFRadialDisplay and a RadarDisplay object.

"""
print __doc__

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import netCDF4
import matplotlib.pyplot as plt
import pyart

filename = 'sgpxsaprrhicmacI5.c0.20110524.015604_NC4.nc'

# create the plot using RadarDisplay (recommended method)
radar = pyart.io.read_cfradial(filename)
radar.metadata['instrument_name'] = 'XSARP'
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure(figsize=[12, 17])
fig.subplots_adjust(hspace=0.4)
xlabel = 'Distance from radar (km)'
ylabel = 'Height agl (km)'
colorbar_label = 'Hz. Eq. Refl. Fac. (dBZ)'
nplots = radar.nsweeps

for snum in radar.sweep_number['data']:

    fixed_angle = radar.fixed_angle['data'][snum]
    title = 'HSRHI Az=%.3f' % (fixed_angle)
    ax = fig.add_subplot(nplots, 1, snum+1)
    display.plot('reflectivity_horizontal', snum, vmin=-20, vmax=20,
                 mask_outside=False, title=title,
                 axislabels=(xlabel, ylabel),
                 colorbar_label=colorbar_label, ax=ax)
    display.set_limits(ylim=[0, 15], ax=ax)

figure_title = 'Time: ' + display.time_begin.isoformat() + 'Z'
fig.text(0.35, 0.92, figure_title)

plt.show()

# create the plot using CFRadialDisplay
dataset = netCDF4.Dataset(filename, 'r')
display = pyart.graph.CFRadialDisplay(dataset)

fig = plt.figure(figsize=[12, 17])
fig.subplots_adjust(hspace=0.4)
xlabel = 'Distance from radar (km)'
ylabel = 'Height agl (km)'
colorbar_label = 'Hz. Eq. Refl. Fac. (dBZ)'
nplots = len(dataset.variables['sweep_number'])

for snum in dataset.variables['sweep_number']:

    title = 'HSRHI Az=%.3f' % dataset.variables['fixed_angle'][snum]
    ax = fig.add_subplot(nplots, 1, snum+1)
    display.plot_rhi('reflectivity_horizontal', snum, vmin=-20, vmax=20,
                     mask_outside=False, title=title,
                     axislabels=(xlabel, ylabel),
                     colorbar_label=colorbar_label, ax=ax)
    display.set_limits(ylim=[0, 15], ax=ax)

figure_title = 'Time: ' + display.time_begin.isoformat() + 'Z'
fig.text(0.35, 0.92, figure_title)

plt.show()



""" Unit Tests for Py-ART's io/mdv.py module. """

from datetime import datetime

import numpy as np
from numpy.ma.core import MaskedArray

import pyart

############################################
# read_mdv tests (verify radar attributes) #
############################################

# read in the sample file and create a a Radar object
radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)


# time attribute
def test_time():
    assert 'comment' in radar.time.keys()
    assert 'long_name' in radar.time.keys()
    assert 'standard_name' in radar.time.keys()
    assert 'units' in radar.time.keys()
    assert 'calendar' in radar.time.keys()
    assert 'data' in radar.time.keys()
    assert radar.time['units'] == 'seconds since 2011-05-20T11:01:00Z'
    assert radar.time['data'].shape == (360, )
    assert round(radar.time['data'][200]) == 187.


# range attribute
def test_range():
    assert 'long_name' in radar.range
    assert 'standard_name' in radar.range
    assert 'meters_to_center_of_first_gate' in radar.range
    assert 'meters_between_gates' in radar.range
    assert 'units' in radar.range
    assert 'data' in radar.range
    assert 'spacing_is_constant' in radar.range
    assert radar.range['data'].shape == (110, )
    assert round(radar.range['data'][0]) == 118.0


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert 'instrument_name' in radar.metadata
    assert 'source' in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == 'ppi'


# latitude attribute
def test_latitude():
    assert 'data' in radar.latitude
    assert 'standard_name' in radar.latitude
    assert 'units' in radar.latitude
    assert radar.latitude['data'].shape == (1, )
    assert round(radar.latitude['data']) == 37.0


# longitude attribute
def test_longitude():
    assert 'data' in radar.longitude
    assert 'standard_name' in radar.longitude
    assert 'units' in radar.longitude
    assert radar.longitude['data'].shape == (1, )
    assert round(radar.longitude['data']) == -97.0


# altitude attribute
def test_altitude():
    assert 'data' in radar.altitude
    assert 'standard_name' in radar.altitude
    assert 'units' in radar.altitude
    assert 'positive' in radar.altitude
    assert radar.altitude['data'].shape == (1, )
    assert round(radar.altitude['data']) == 328.0


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert 'standard_name' in radar.sweep_number
    assert np.all(radar.sweep_number['data'] == range(1))


# sweep_mode attribute
def test_sweep_mode():
    assert 'standard_name' in radar.sweep_mode
    assert radar.sweep_mode['data'].shape == (1, )
    assert np.all(radar.sweep_mode['data'] == ['azimuth_surveillance'])


# fixed_angle attribute
def test_fixed_angle():
    assert 'standard_name' in radar.fixed_angle
    assert 'units' in radar.fixed_angle
    assert radar.fixed_angle['data'].shape == (1, )
    assert round(radar.fixed_angle['data'][0], 2) == 0.75


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert 'long_name' in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index['data'].shape == (1, )
    assert round(radar.sweep_start_ray_index['data'][0]) == 0.0


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert 'long_name' in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index['data'].shape == (1, )
    assert round(radar.sweep_end_ray_index['data'][0]) == 359.0


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert 'standard_name' in radar.azimuth
    assert 'long_name' in radar.azimuth
    assert 'units' in radar.azimuth
    assert 'axis' in radar.azimuth
    assert round(radar.azimuth['data'][0]) == 0.0
    assert round(radar.azimuth['data'][10]) == 10.0


# elevation attribute
def test_elevation():
    assert 'standard_name' in radar.elevation
    assert 'long_name' in radar.azimuth
    assert 'units' in radar.elevation
    assert 'axis' in radar.elevation
    assert radar.elevation['data'].shape == (360, )
    assert round(radar.elevation['data'][0], 2) == 0.75


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
def test_instument_parameters():
    # instrument_parameter sub-convention
    keys = ['prt', 'unambiguous_range', 'prt_mode', 'nyquist_velocity']
    for k in keys:
        description = 'instrument_parameters: %s' % k
        check_instrument_parameter.description = description
        yield check_instrument_parameter, k


def check_instrument_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic['meta_group'] == 'instrument_parameters'


# radar_parameters attribute
def test_radar_parameters():
    # radar_parameter sub-convention
    keys = ['radar_beam_width_h', 'radar_beam_width_v']
    for k in keys:
        description = 'radar_parameters: %s' % k
        check_radar_parameter.description = description
        yield check_radar_parameter, k


def check_radar_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic['meta_group'] == 'radar_parameters'


# radar_calibration attribute
def test_radar_calibration():
    assert radar.radar_calibration is None


# ngates attribute
def test_ngates():
    assert radar.ngates == 110


# nrays attribute
def test_nrays():
    assert radar.nrays == 360


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 1


####################
# fields attribute #
####################


def test_field_dics():
    fields = ['reflectivity', ]
    for field in fields:
        description = "field : %s, dictionary" % field
        check_field_dic.description = description
        yield check_field_dic, field


def check_field_dic(field):
    """ Check that the required keys are present in a field dictionary. """
    assert 'standard_name' in radar.fields[field]
    assert 'units' in radar.fields[field]
    assert '_FillValue' in radar.fields[field]
    assert 'coordinates' in radar.fields[field]


def test_field_shapes():
    fields = ['reflectivity', ]
    for field in fields:
        description = "field : %s, shape" % field
        check_field_shape.description = description
        yield check_field_shape, field


def check_field_shape(field):
    assert radar.fields[field]['data'].shape == (360, 110)


def test_field_types():
    fields = {'reflectivity': MaskedArray, }
    for field, field_type in fields.iteritems():
        description = "field : %s, type" % field
        check_field_type.description = description
        yield check_field_type, field, field_type


def check_field_type(field, field_type):
    assert type(radar.fields[field]['data']) is field_type


def test_field_first_points():
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    fields = {'reflectivity': 24.0}
    for field, field_value in fields.iteritems():
        description = "field : %s, first point" % field
        check_field_first_point.description = description
        yield check_field_first_point, field, field_value


def check_field_first_point(field, value):
    assert round(radar.fields[field]['data'][0, 0]) == value


#################
# MdvFile tests #
#################


mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_PPI_FILE)


def test_master_header():
    # test the master header
    ref_master_header = {
        'chunk_hdr_offset': 2464,
        'data_collection_type': 0,
        'data_dimension': 0,
        'data_ordering': 0,
        'data_set_info': 'MDV radar volume file created by Dsr2Vol.',
        'data_set_name': 'C-SAPR',
        'data_set_source': 'ARM SGP C-SAPR',
        'field_grids_differ': 0,
        'field_hdr_offset': 1024,
        'grid_orientation': 1,
        'index_number': 611,
        'native_vlevel_type': 9,
        'nchunks': 3,
        'nfields': 1,
        'ngates': 110,
        'nrays': 360,
        'nsweeps': 1,
        'num_data_times': 1,
        'record_len1': 1016,
        'record_len2': 1016,
        'revision_number': 1,
        'sensor_alt': 0.32760000228881836,
        'sensor_lat': 36.79615783691406,
        'sensor_lon': -97.45054626464844,
        'struct_id': 14142,
        'time_begin': 1305889260,
        'time_centroid': 1305889595,
        'time_end': 1305889595,
        'time_expire': 1305890265,
        'time_gen': 1305889595,
        'time_written': 1305889668,
        'unused_fl3212': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0),
        'unused_si325': (0, 0, 0, 0, 0),
        'user_data': 0,
        'user_data_fl326': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        'user_data_si328': (0, 0, 0, 0, 0, 0, 0, 0),
        'user_time': 0,
        'vlevel_hdr_offset': 1440,
        'vlevel_included': 1,
        'vlevel_type': 9}

    for k, v in ref_master_header.iteritems():
        print "checking key:", k
        assert mdvfile.master_header[k] == v


def test_field_header():
    # test a few of the field_header
    assert len(mdvfile.field_headers) == 1
    assert mdvfile.field_headers[0]['bad_data_value'] == 0.0
    assert mdvfile.field_headers[0]['dz_constant'] == 0
    assert mdvfile.field_headers[0]['encoding_type'] == 2
    assert mdvfile.field_headers[0]['field_name'] == "DBZ_F"
    assert mdvfile.field_headers[0]['nsweeps'] == 1


def test_vlevel_headers():
    # test the vlevel_headers
    assert len(mdvfile.vlevel_headers) == 1
    assert mdvfile.vlevel_headers[0]['record_len1'] == 1016
    assert mdvfile.vlevel_headers[0]['record_len2'] == 1016
    assert mdvfile.vlevel_headers[0]['unused_fl32'] == (0.0, 0.0, 0.0,
                                                        0.0, 0.0)
    assert mdvfile.vlevel_headers[0]['unused_si32'] == (0, 0, 0, 0)
    assert mdvfile.vlevel_headers[0]['struct_id'] == 14144
    assert len(mdvfile.vlevel_headers[0]['level']) == 122
    assert mdvfile.vlevel_headers[0]['level'][0] == 0.75
    assert mdvfile.vlevel_headers[0]['level'][10] == 11.699999809265137
    assert mdvfile.vlevel_headers[0]['level'][17] == 0.0
    assert len(mdvfile.vlevel_headers[0]['type']) == 122
    assert mdvfile.vlevel_headers[0]['type'][0:17] == (9,) * 17
    assert mdvfile.vlevel_headers[0]['type'][18] == 0


def test_chunk_headers():
    # test chunk_headers
    assert len(mdvfile.chunk_headers) == 3
    ref_chunk_header_0 = {
        'chunk_data_offset': 68580,
        'chunk_id': 3,
        'info': 'DsRadar params',
        'record_len1': 504,
        'record_len2': 504,
        'size': 240,
        'struct_id': 14145,
        'unused_si32': (0, 0)}
    for k, v in ref_chunk_header_0.iteritems():
        print "checking key:", k
        assert mdvfile.chunk_headers[0][k] == v


def test_calib_info():
    # test calib_info
    ref_calib_info = {
        'I0_h_co_dbm': 0.0,
        'I0_h_cx_dbm': 0.0,
        'I0_v_co_dbm': 0.0,
        'I0_v_cx_dbm': 0.0,
        'antenna_gain_h_db': 44.5,
        'antenna_gain_v_db': 44.5,
        'beamwidth_h_deg': 1.0,
        'beamwidth_v_deg': 1.0,
        'coupler_fwd_loss_h_db': 50.0,
        'coupler_fwd_loss_v_db': 50.0,
        'day': 15,
        'filter_loss_db': 1.600000023841858,
        'hour': 18,
        'ldr_h_bias_db': 0.0,
        'ldr_v_bias_db': 0.0,
        'minute': 49,
        'month': 4,
        'noise_h_co_dbm': -73.5886001586914,
        'noise_h_cx_dbm': -73.5886001586914,
        'noise_source_h_dbm': -9999.0,
        'noise_source_v_dbm': -9999.0,
        'noise_v_co_dbm': -73.13050079345703,
        'noise_v_cx_dbm': -73.13050079345703,
        'power_meas_loss_h_db': 4.96999979019165,
        'power_meas_loss_v_db': 5.840000152587891,
        'pulse_width_us': 0.800000011920929,
        'radar_constant_h_db': 73.25920104980469,
        'radar_constant_v_db': 73.25920104980469,
        'radar_name': 'C-SAPR',
        'rx_gain_h_co_dbm': 33.31079864501953,
        'rx_gain_h_cx_dbm': 33.31079864501953,
        'rx_gain_v_co_dbm': 34.24689865112305,
        'rx_gain_v_cx_dbm': 34.24689865112305,
        'rx_slope_h_co_db': 1.0030399560928345,
        'rx_slope_h_cx_db': 1.0030399560928345,
        'rx_slope_v_co_db': 0.9963300228118896,
        'rx_slope_v_cx_db': 0.9963300228118896,
        'second': 36,
        'spare': (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                  0.0, 0.0),
        'sun_h_co_dbm': -9999.0,
        'sun_h_cx_dbm': -9999.0,
        'sun_v_co_dbm': -9999.0,
        'sun_v_cx_dbm': -9999.0,
        'system_phidp_deg': -60.0,
        'test_pulse_h_dbm': -9999.0,
        'test_pulse_v_dbm': -9999.0,
        'twoway_radome_loss_h_db': 0.5,
        'twoway_radome_loss_v_db': 0.5,
        'twoway_waveguide_loss_h_db': 4.0,
        'twoway_waveguide_loss_v_db': 4.0,
        'wavelength_cm': 5.329639911651611,
        'xmit_power_h_dbm': 83.6500015258789,
        'xmit_power_v_dbm': 83.6500015258789,
        'year': 2011,
        'zdr_bias_db': -1.2799999713897705,
        'zh1km_co_dbz': -33.63949966430664,
        'zh1km_cx_dbz': -33.63949966430664,
        'zv1km_co_dbz': -34.11470031738281,
        'zv1km_cx_dbz': -34.11470031738281}

    for k, v in ref_calib_info.iteritems():
        print "checking key:", k
        assert mdvfile.calib_info[k] == v


def test_radar_info():

    # check radar_info
    ref_radar_info = {
        'altitude_km': 0.32760000228881836,
        'antenna_gain_db': 44.5,
        'field_flag': 0,
        'follow_mode': 1,
        'gate_spacing_km': 0.11991698294878006,
        'horiz_beam_width_deg': 1.0,
        'latitude_deg': 36.79615783691406,
        'longitude_deg': -97.45054626464844,
        'measXmitPowerDbmH_dbm': 82.7537612915039,
        'measXmitPowerDbmV_dbm': 82.65692138671875,
        'nfields': 1,
        'nfields_current': 0,
        'ngates': 110,
        'polarization': 5,
        'prf_hz': 1239.9993896484375,
        'prf_mode': 1,
        'prt2_s': 0.0,
        'prt_s': 0.0008064520079642534,
        'pulse_width_us': 0.800000011920929,
        'radar_constant': 73.25920104980469,
        'radar_id': 0,
        'radar_name': 'C-SAPR',
        'radar_type': 0,
        'receiver_gain_db': 33.31079864501953,
        'receiver_mds_dbm': -106.89939880371094,
        'samples_per_beam': 128,
        'scan_mode': 8,
        'scan_type': 0,
        'scan_type_name': 'Default_s',
        'spare_floats': (0.0, 0.0, 0.0, 0.0),
        'spare_ints': (0, 0),
        'start_range_km': 0.11787839233875275,
        'system_gain_db': 77.81079864501953,
        'unambig_range_km': 117.99627685546875,
        'unambig_vel_mps': 16.52467918395996,
        'vert_beam_width_deg': 1.0,
        'wavelength_cm': 5.330544471740723,
        'xmit_peak_pwr_watts': 231739552.0}

    for k, v in ref_radar_info.iteritems():
        print "checking key:", k
        assert mdvfile.radar_info[k] == v


def test_geometry():
    # test geometry attributes
    assert np.all(mdvfile.az_deg == np.arange(360))
    assert len(mdvfile.range_km) == 110
    assert round(mdvfile.range_km[10], 2) == 1.32
    assert len(mdvfile.el_deg) == 1
    assert mdvfile.el_deg[0] == 0.75


def test_mdv_time():
    # test times attribute
    assert mdvfile.times['time_begin'] == datetime(2011, 5, 20, 11, 1)
    assert mdvfile.times['time_centroid'] == datetime(2011, 5, 20, 11, 6, 35)
    assert mdvfile.times['time_end'] == datetime(2011, 5, 20, 11, 6, 35)


def test_cart():
    # test cartography information
    assert mdvfile.scan_type == 'ppi'
    assert mdvfile.carts['x'].shape == (1, 360, 110)
    assert mdvfile.carts['y'].shape == (1, 360, 110)
    assert mdvfile.carts['z'].shape == (1, 360, 110)
    assert round(mdvfile.carts['x'][0, 1, 2], 2) == 6.24
    assert round(mdvfile.carts['y'][0, 1, 2], 2) == 357.63
    assert round(mdvfile.carts['z'][0, 1, 2], 2) == 4.69


def test_fields():
    # test fields
    ref_fields = ['DBZ_F', ]
    for a, b in zip(ref_fields, mdvfile.fields):
        assert a == b


def test_fileptr():
    # test fileptr
    assert type(mdvfile.fileptr) == file


def test_read_one_field():

    mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_PPI_FILE)
    # extract a field
    assert hasattr(mdvfile, 'DBZ_F') is False
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert round(sweeps[0, 1, 2], 2) == 13.19
    assert hasattr(mdvfile, 'DBZ_F') is True

    # reread, should be faster
    sweeps = mdvfile.read_a_field(0)
    assert sweeps.shape == (1, 360, 110)
    assert round(sweeps[0, 1, 2], 2) == 13.19
    assert hasattr(mdvfile, 'DBZ_F') is True

    delattr(mdvfile, 'DBZ_F')


def test_read_all_fields():

    mdvfile = pyart.io.mdv.MdvFile(pyart.testing.MDV_PPI_FILE)
    # read all fields
    assert hasattr(mdvfile, 'DBZ_F') is False
    mdvfile.read_all_fields()
    assert hasattr(mdvfile, 'DBZ_F') is True
    assert round(mdvfile.DBZ_F[0, 1, 2], 2) == 13.19


def test_read_all_fields_on_creation():
    mdvfile2 = pyart.io.mdv.MdvFile(pyart.testing.MDV_PPI_FILE,
                                    read_fields=True)
    assert hasattr(mdvfile2, 'DBZ_F') is True
    assert round(mdvfile2.DBZ_F[0, 1, 2], 2) == 13.19
    mdvfile2.close()


#############
# RHI tests #
#############

RADAR_RHI = pyart.io.read_mdv(pyart.testing.MDV_RHI_FILE)


# nsweeps attribute
def test_rhi_nsweeps():
    assert RADAR_RHI.nsweeps == 1


# sweep_number attribute
def test_rhi_sweep_number():
    assert 'standard_name' in RADAR_RHI.sweep_number
    assert np.all(RADAR_RHI.sweep_number['data'] == range(1))


# sweep_mode attribute
def test_rhi_sweep_mode():
    assert 'standard_name' in RADAR_RHI.sweep_mode
    assert RADAR_RHI.sweep_mode['data'].shape == (1, )
    assert np.all(RADAR_RHI.sweep_mode['data'] == ['rhi'])


# fixed_angle attribute
def test_rhi_fixed_angle():
    assert 'standard_name' in RADAR_RHI.fixed_angle
    assert 'units' in RADAR_RHI.fixed_angle
    assert RADAR_RHI.fixed_angle['data'].shape == (1, )
    assert round(RADAR_RHI.fixed_angle['data'][0], 2) == 189.


# sweep_start_ray_index attribute
def test_rhi_sweep_start_ray_index():
    assert 'long_name' in RADAR_RHI.sweep_start_ray_index
    assert RADAR_RHI.sweep_start_ray_index['data'].shape == (1, )
    assert round(RADAR_RHI.sweep_start_ray_index['data'][0]) == 0.0


# sweep_end_ray_index attribute
def test_rhi_sweep_end_ray_index():
    assert 'long_name' in RADAR_RHI.sweep_end_ray_index
    assert RADAR_RHI.sweep_end_ray_index['data'].shape == (1, )
    assert round(RADAR_RHI.sweep_end_ray_index['data'][0]) == 282.0


# azimuth attribute
def test_rhi_azimuth():
    assert 'standard_name' in RADAR_RHI.azimuth
    assert 'long_name' in RADAR_RHI.azimuth
    assert 'units' in RADAR_RHI.azimuth
    assert 'axis' in RADAR_RHI.azimuth
    assert round(RADAR_RHI.azimuth['data'][0]) == 189.0
    assert round(RADAR_RHI.azimuth['data'][10]) == 189.0


# elevation attribute
def test_rhi_elevation():
    assert 'standard_name' in RADAR_RHI.elevation
    assert 'long_name' in RADAR_RHI.azimuth
    assert 'units' in RADAR_RHI.elevation
    assert 'axis' in RADAR_RHI.elevation
    assert RADAR_RHI.elevation['data'].shape == (283, )
    assert round(RADAR_RHI.elevation['data'][0], 2) == 19.6


def test_open_from_file_obj():
    fh = open(pyart.testing.MDV_PPI_FILE, 'rb')
    radar = pyart.io.read_mdv(pyart.testing.MDV_PPI_FILE)
    fh.close()



"""
pyart.aux_io.radx
=================

Reading files using Radx to first convert the file to Cf.Radial format

.. autosummary::
    :toctree: generated/

    read_radx

"""

import os
import tempfile
import subprocess

from ..io.cfradial import read_cfradial


def read_radx(filename):
    """
    Read a file by first converting it to Cf/Radial using RadxConvert.

    Parameters
    ----------
    filename : str
        Name of file to read using RadxConvert.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    tmpfile = tempfile.mkstemp(suffix='.nc', dir='.')[1]
    head, tail = os.path.split(tmpfile)
    try:
        subprocess.check_call(
            ['RadxConvert', '-const_ngates',
             '-outdir', head, '-outname', tail, '-f', filename])
        radar = read_cfradial(tmpfile)
    finally:
        os.remove(tmpfile)
    return radar



"""
========================================
Testing Utilities (:mod:`pyart.testing`)
========================================

.. currentmodule:: pyart.testing

Py-ART comes with a number of utilities helpful when writing and running
unit tests.

.. autosummary::
    :toctree: generated/

    make_empty_ppi_radar
    make_target_radar
    make_single_ray_radar
    make_velocity_aliased_radar
    make_empty_grid
    make_target_grid

"""

from .sample_files import MDV_PPI_FILE, MDV_RHI_FILE
from .sample_files import CFRADIAL_PPI_FILE, CFRADIAL_RHI_FILE
from .sample_files import CHL_RHI_FILE
from .sample_files import SIGMET_PPI_FILE, SIGMET_RHI_FILE
from .sample_files import INTERP_SOUNDE_FILE
from .sample_files import NEXRAD_ARCHIVE_FILE, NEXRAD_CDM_FILE
from .sample_files import NEXRAD_ARCHIVE_COMPRESSED_FILE
from .sample_objects import make_empty_ppi_radar, make_target_radar
from .sample_objects import make_single_ray_radar, make_velocity_aliased_radar
from .sample_objects import make_empty_grid
from .sample_objects import make_target_grid, make_storm_grid
from .sample_objects import make_empty_rhi_radar
from .sample_objects import make_velocity_aliased_rhi_radar

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.aux_io.read_gamic
=======================

Utilities for reading gamic hdf5 files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    _GAMICFileHelper

.. autosummary::
    :toctree: generated/

    read_gamic
    _get_instrument_params
    _avg_radial_angles
    _prt_mode_from_unfolding
    _get_gamic_sweep_data


"""

# TODO to move out of aux_io namespace:
# * unit tests - need small sample files.
# * auto-detect file type with pyart.io.read function
# * move to pyart.io namespace

import datetime
import warnings

import h5py
import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str
from ..core.radar import Radar


LIGHT_SPEED = 2.99792458e8  # speed of light in meters per second


def read_gamic(filename, field_names=None, additional_metadata=None,
               file_field_names=False, exclude_fields=None,
               valid_range_from_file=True, units_from_file=True):
    """
    Read a GAMIC hdf5 file.

    Parameters
    ----------
    filename : str
        Name of GAMIC HDF5 file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    valid_range_from_file : bool, optional
        True to extract valid range (valid_min and valid_max) for all
        field from the file when they are present.  False will not extract
        these parameters.
    units_from_file : bool, optional
        True to extract the units for all fields from the file when available.
        False will not extract units using the default units for the fields.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # create metadata retrieval object
    filemetadata = FileMetadata('gamic', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    # Open HDF5 file and get handle
    hfile = h5py.File(filename, 'r')
    ghelp = _GAMICFileHelper(hfile)

    # verify that scan0 ... scan[nsweeps-1] are present in file
    for scan in ghelp.scans:
        assert scan in hfile

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = ghelp.where_attr('lat', 'float64')
    longitude['data'] = ghelp.where_attr('lon', 'float64')
    altitude['data'] = ghelp.where_attr('height', 'float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'GAMIC-HDF5'
    what_mapping = {
        'version': 'gamic_version', 'sets_scheduled': 'sets_scheduled',
        'object': 'gamic_object', 'date': 'gamic_date'}
    for hkey, mkey in what_mapping.items():
        if hkey in hfile['what'].attrs:
            metadata[mkey] = hfile['what'].attrs[hkey]

    how_keys = ['software', 'template_name', 'site_name', 'host_name',
                'azimuth_beam', 'elevation_beam', 'sdp_name', 'sw_version',
                'sdp_version', 'simulated']
    for key in how_keys:
        if key in hfile['how'].attrs:
            metadata[key] = hfile['how'].attrs[key]

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_start_ray_index['data'] = ghelp.start_ray.astype('int32')
    sweep_end_ray_index['data'] = ghelp.end_ray.astype('int32')

    # sweep number
    sweep_number = filemetadata('sweep_number')
    try:
        sweep_number['data'] = ghelp.what_attrs('set_idx', 'int32')
    except KeyError:
        sweep_number['data'] = np.arange(len(ghelp.scans), dtype='int32')

    # sweep_type
    scan_type = hfile['scan0']['what'].attrs['scan_type'].lower()
    # check that all scans in the volume are the same type
    for scan in ghelp.scans:
        if hfile[scan]['what'].attrs['scan_type'].lower() != scan_type:
            raise NotImplementedError('Mixed scan_type volume in scan:' + scan)
    if scan_type not in ['ppi', 'rhi']:
        message = "Unknown scan type: %s, reading as RHI scans." % (scan_type)
        warnings.warn(message)
        scan_type = 'rhi'

    # sweep_mode, fixed_angle
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    if scan_type == 'rhi':
        sweep_mode['data'] = np.array(ghelp.nsweeps * ['rhi'])
        fixed_angle['data'] = ghelp.how_attrs('azimuth', 'float32')
    elif scan_type == 'ppi':
        sweep_mode['data'] = np.array(ghelp.nsweeps * ['azimuth_surveillance'])
        fixed_angle['data'] = ghelp.how_attrs('elevation', 'float32')

    # time
    time = filemetadata('time')
    t_data = ghelp.ray_header('timestamp', 'int64')
    start_epoch = t_data[0] // 1.e6     # truncate to second resolution
    start_time = datetime.datetime.utcfromtimestamp(start_epoch)
    time['units'] = make_time_unit_str(start_time)
    time['data'] = ((t_data - start_epoch * 1.e6) / 1.e6).astype('float64')

    # range
    _range = filemetadata('range')
    ngates = int(hfile['/scan0/how'].attrs['bin_count'])
    range_start = float(hfile['/scan0/how'].attrs['range_start'])
    range_step = float(hfile['/scan0/how'].attrs['range_step'])
    # range_step may need to be scaled by range_samples
    # XXX This gives distances to start of gates not center, this matches
    # Radx but may be incorrect, add range_step / 2. for center
    _range['data'] = (np.arange(ngates, dtype='float32') * range_step +
                      range_start)
    _range['meters_to_center_of_first_gate'] = range_start
    _range['meters_between_gates'] = range_step

    # elevation
    elevation = filemetadata('elevation')
    start_angle = ghelp.ray_header('elevation_start', 'float32')
    stop_angle = ghelp.ray_header('elevation_stop', 'float32')
    elevation['data'] = _avg_radial_angles(start_angle, stop_angle)

    # azimuth
    azimuth = filemetadata('azimuth')
    start_angle = ghelp.ray_header('azimuth_start', 'float32')
    stop_angle = ghelp.ray_header('azimuth_stop', 'float32')
    azimuth['data'] = _avg_radial_angles(start_angle, stop_angle) % 360.

    # fields
    fields = {}
    h_keys = [k for k in hfile['/scan0'] if k.startswith('moment_')]
    h_fields = [hfile['/scan0'][k].attrs['moment'] for k in h_keys]

    for h_field, h_key in zip(h_fields, h_keys):

        field_name = filemetadata.get_field_name(h_field)
        if field_name is None:
            continue

        field_dic = filemetadata(field_name)
        field_dic['data'] = ghelp.moment_data(h_key, 'float32')
        field_dic['_FillValue'] = get_fillvalue()

        if valid_range_from_file:
            try:
                valid_min = hfile['/scan0'][h_key].attrs['dyn_range_min']
                valid_max = hfile['/scan0'][h_key].attrs['dyn_range_max']
                field_dic['valid_min'] = valid_min
                field_dic['valid_max'] = valid_max
            except:
                pass

        if units_from_file:
            try:
                field_dic['units'] = hfile['/scan0'][h_key].attrs['unit']
            except:
                pass
        fields[field_name] = field_dic

    # ray_angle_res
    ray_angle_res = filemetadata('ray_angle_res')
    ray_angle_res['data'] = ghelp.how_attrs('angle_step', 'float32')

    # rays_are_indexed
    rays_are_indexed = filemetadata('rays_are_indexed')
    rays_are_indexed['data'] = np.array(
        [['false', 'true'][i] for i in ghelp.how_attrs('angle_sync', 'uint8')])

    # target_scan_rate
    target_scan_rate = filemetadata('target_scan_rate')
    target_scan_rate['data'] = ghelp.how_attrs('scan_speed', 'float32')

    # scan_rate
    scan_rate = filemetadata('scan_rate')
    if scan_type == 'ppi':
        scan_rate['data'] = ghelp.ray_header('az_speed', 'float32')
    elif scan_type == 'rhi':
        scan_rate['data'] = ghelp.ray_header('el_speed', 'float32')
    else:
        scan_rate = None

    # instrument_parameters
    instrument_parameters = _get_instrument_params(ghelp, filemetadata)

    hfile.close()

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        ray_angle_res=ray_angle_res,
        rays_are_indexed=rays_are_indexed,
        scan_rate=scan_rate,
        target_scan_rate=target_scan_rate)


def _get_instrument_params(ghelp, filemetadata):
    """ Return a dictionary containing instrument parameters. """

    instrument_params = {}

    dic = filemetadata('frequency')
    dic['data'] = np.array(
        [LIGHT_SPEED / ghelp.hfile['scan0']['how'].attrs['radar_wave_length']],
        dtype='float32')
    instrument_params['frequency'] = dic

    dic = filemetadata('radar_beam_width_h')
    dic['data'] = ghelp.how_attr('azimuth_beam', 'float32')
    instrument_params['radar_beam_width_h'] = dic

    dic = filemetadata('radar_beam_width_v')
    dic['data'] = ghelp.how_attr('elevation_beam', 'float32')
    instrument_params['radar_beam_width_v'] = dic

    dic = filemetadata('pulse_width')
    dic['data'] = ghelp.sweep_expand(
        ghelp.how_attrs('pulse_width_us', 'float32') * 1e-6)
    instrument_params['pulse_width'] = dic

    dic = filemetadata('prt')
    dic['data'] = ghelp.sweep_expand(1. / ghelp.how_attrs('PRF', 'float32'))
    instrument_params['prt'] = dic

    unfolding = ghelp.how_attrs('unfolding', 'int32')
    dic = filemetadata('prt_mode')
    dic['data'] = np.array([_prt_mode_from_unfolding(i) for i in unfolding])
    instrument_params['prt_mode'] = dic

    dic = filemetadata('prt_ratio')
    dic['data'] = ghelp.sweep_expand(
        [[1, 2./3., 3./4., 4./5.][i] for i in unfolding])
    instrument_params['prt_ratio'] = dic

    dic = filemetadata('unambiguous_range')
    dic['data'] = ghelp.sweep_expand(ghelp.how_attrs('range', 'float32'))
    instrument_params['unambiguous_range'] = dic

    dic = filemetadata('nyquist_velocity')
    dic['data'] = ghelp.sweep_expand(ghelp.how_ext_attrs('nyquist_velocity'))
    instrument_params['nyquist_velocity'] = dic

    dic = filemetadata('n_samples')
    dic['data'] = ghelp.sweep_expand(
        ghelp.how_attrs('range_samples', 'int32') *
        ghelp.how_attrs('time_samples', 'int32'), dtype='int32')
    instrument_params['n_samples'] = dic

    return instrument_params


class _GAMICFileHelper(object):
    """
    A class to help read GAMIC files.

    This class has methods which performs looping routines which are
    useful when reading GAMIC files.
    """

    def __init__(self, hfile):
        """ initialize object. """

        self.hfile = hfile
        self.nsweeps = hfile['what'].attrs['sets']
        self.scans = ['scan%i' % (i) for i in range(self.nsweeps)]
        self.rays_per_sweep = self.how_attrs('ray_count', 'int32')
        self.total_rays = sum(self.rays_per_sweep)
        # starting and ending ray for each sweep
        self.start_ray = np.cumsum(np.append([0], self.rays_per_sweep[:-1]))
        self.end_ray = np.cumsum(self.rays_per_sweep) - 1

        return

    # attribute look up
    def where_attr(self, attr, dtype):
        """ Return an array containing a attribute from the where group. """
        return np.array([self.hfile['where'].attrs[attr]], dtype=dtype)

    def how_attr(self, attr, dtype):
        """ Return an array containing a attribute from the how group. """
        return np.array([self.hfile['how'].attrs[attr]], dtype=dtype)

    # scan/sweep based attribute lookup
    def how_attrs(self, attr, dtype):
        """ Return an array of an attribute for each scan's how group. """
        return np.array([self.hfile[s]['how'].attrs[attr]
                         for s in self.scans], dtype=dtype)

    def how_ext_attrs(self, attr):
        """
        Return a list of an attribute in each scan's how/extended group.
        """
        return [float(self.hfile[s]['how']['extended'].attrs[attr])
                for s in self.scans]

    def what_attrs(self, attr, dtype):
        """ Return a list of an attribute for each scan's what group. """
        return np.array([self.hfile[s]['what'].attrs[attr]
                         for s in self.scans], dtype=dtype)

    # misc looping
    def ray_header(self, field, dtype):
        """ Return an array containing a ray_header field for each sweep. """
        data = np.empty((self.total_rays, ), dtype=dtype)
        for scan, start, end in zip(self.scans, self.start_ray, self.end_ray):
            data[start:end+1] = self.hfile[scan]['ray_header'][field]
        return data

    def moment_data(self, moment, dtype):
        """ Read in moment data from all sweeps. """
        ngates = int(self.hfile['/scan0/how'].attrs['bin_count'])
        data = np.ma.zeros((self.total_rays, ngates), dtype=dtype)
        data[:] = np.ma.masked      # volume data initially all masked
        for scan, start, end in zip(self.scans, self.start_ray, self.end_ray):
            # read in sweep data if field exists in scan.
            if moment in self.hfile[scan]:
                sweep_data = _get_gamic_sweep_data(self.hfile[scan][moment])
                data[start:end+1, :sweep_data.shape[1]] = sweep_data[:]
        return data

    def sweep_expand(self, arr, dtype='float32'):
        """ Expand an sweep indexed array to be ray indexed """
        return np.repeat(arr, self.rays_per_sweep).astype(dtype)


def _avg_radial_angles(angle1, angle2):
    """ Return the average angle between two radial angles. """
    return np.angle(
        (np.exp(1.j*np.deg2rad(angle1)) +
         np.exp(1.j*np.deg2rad(angle2))) / 2., deg=True)


def _prt_mode_from_unfolding(unfolding):
    """ Return 'fixed' or 'staggered' depending on unfolding flag """
    if unfolding == 0:
        return 'fixed'
    else:
        return 'staggered'


def _get_gamic_sweep_data(group):
    """ Get GAMIC HDF5 sweep data from an HDF5 group. """
    dyn_range_min = group.attrs['dyn_range_min']
    dyn_range_max = group.attrs['dyn_range_max']
    raw_data = group[:]
    fmt = group.attrs['format']
    if fmt == 'UV16':
        # unsigned 16-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint16
        scale = (dyn_range_max - dyn_range_min) / 65535.
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype='float32')
    elif fmt == 'UV8':
        # unsigned 8-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint8
        scale = (dyn_range_max - dyn_range_min) / 255.
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype='float32')
    else:
        raise NotImplementedError('GAMIC data format: %s', fmt)
    return sweep_data




# THIS FILE IS GENERATED FROM PYART SETUP.PY
short_version = '1.3.0'
version = '1.3.0'
full_version = '1.3.0'
git_revision = 'a15c4e4e7fe28dca7ae0835cdb059081830d8ee0'
release = True

if not release:
    version = full_version



"""
pyart.io.nexrad_archive
=======================

Functions for reading NEXRAD Level II Archive files.

.. autosummary::
    :toctree: generated/

    read_nexrad_archive

"""

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str
from .nexrad_level2 import NEXRADLevel2File


def read_nexrad_archive(filename, field_names=None, additional_metadata=None,
                        file_field_names=False, exclude_fields=None,
                        bzip=None):
    """
    Read a NEXRAD Level 2 Archive file.

    Parameters
    ----------
    filename : str
        Filename of NEXRAD Level 2 Archive file.  The files hosted by
        at the NOAA National Climate Data Center [1]_ as well as on the
        UCAR THREDDS Data Server [2]_ have been tested.  Other NEXRAD
        Level 2 Archive files may or may not work.  Message type 1 file
        at not yet supported, only message type 31.
    field_names : dict, optional
        Dictionary mapping NEXRAD moments to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        metadata configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the metadata configuration file will be used.
    file_field_names : bool, optional
        True to use the NEXRAD field names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    bzip : bool or None
        True if the file is compressed as a bzip2 file, False otherwise.
        None will examine the filename for a bzip extension.

    Returns
    -------
    radar : Radar
        Radar object containing all moments and sweeps/cuts in the volume.
        Gates not collected are masked in the field data.

    References
    ----------
    .. [1] http://www.ncdc.noaa.gov/
    .. [2] http://thredds.ucar.edu/thredds/catalog.html

    """
    # create metadata retrieval object
    filemetadata = FileMetadata('nexrad_archive', field_names,
                                additional_metadata, file_field_names,
                                exclude_fields)
    # parse bzip parameter
    if bzip is None:
        try:
            if filename.endswith('.bz2') or filename.endswith('bzip2'):
                bzip = True
            else:
                bzip = False
        except:
            bzip = False

    # open the file and retrieve scan information
    nfile = NEXRADLevel2File(filename, bzip)
    scan_info = nfile.scan_info()

    # time
    time = filemetadata('time')
    time_start, _time = nfile.get_times()
    time['data'] = _time
    time['units'] = make_time_unit_str(time_start)

    # range
    scan_max_gates = np.argmax([max(s['ngates']) for s in scan_info])
    i = np.argmax(scan_info[scan_max_gates]['ngates'])
    moment_max_gates = scan_info[scan_max_gates]['moments'][i]
    _range = filemetadata('range')
    _range['data'] = nfile.get_range(scan_max_gates, moment_max_gates)
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    _range['meters_between_gates'] = _range['data'][1] - _range['data'][0]

    # fields
    max_ngates = max([max(s['ngates']) for s in scan_info])
    available_moments = set([])
    for info in scan_info:
        for moment in info['moments']:
            available_moments.add(moment)

    fields = {}
    for moment in available_moments:
        field_name = filemetadata.get_field_name(moment)
        if field_name is None:
            continue
        dic = filemetadata(field_name)
        dic['_FillValue'] = get_fillvalue()
        dic['data'] = nfile.get_data(moment, max_ngates)
        fields[field_name] = dic

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'NEXRAD Level II'

    # scan_type
    scan_type = 'ppi'

    # latitude, longitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    lat, lon, alt = nfile.location()
    latitude['data'] = np.array([lat], dtype='float64')
    longitude['data'] = np.array([lon], dtype='float64')
    altitude['data'] = np.array([alt], dtype='float64')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    nsweeps = int(nfile.nscans)
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')
    sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

    rays_per_scan = [s['nrays'] for s in scan_info]
    sweep_end_ray_index['data'] = np.cumsum(rays_per_scan, dtype='int32') - 1

    rays_per_scan.insert(0, 0)
    sweep_start_ray_index['data'] = np.cumsum(rays_per_scan[:-1],
                                              dtype='int32')

    # azimuth, elevation, fixed_angle
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    fixed_angle = filemetadata('fixed_angle')
    azimuth['data'] = nfile.get_azimuth_angles()
    elevation['data'] = nfile.get_elevation_angles().astype('float32')
    fixed_angle['data'] = nfile.get_target_angles()

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=None)



"""
pyart.correct._common_dealias
=============================

Routines used by multiple dealiasing functions.

.. autosummary::
    :toctree: generated/

    _parse_fields
    _parse_nyquist_vel
    _parse_gatefilter
    _parse_rays_wrap_around

"""

from ..config import get_field_name
from .filters import moment_based_gate_filter, GateFilter


def _parse_fields(vel_field, corr_vel_field):
    """ Parse and return the radar fields for dealiasing. """
    if vel_field is None:
        vel_field = get_field_name('velocity')
    if corr_vel_field is None:
        corr_vel_field = get_field_name('corrected_velocity')
    return vel_field, corr_vel_field


def _parse_nyquist_vel(nyquist_vel, radar):
    """ Parse the nyquist_vel parameter, extract from the radar if needed. """
    if nyquist_vel is None:
        if (radar.instrument_parameters is None) or (
                'nyquist_velocity' not in radar.instrument_parameters):
            message = ('Nyquist velocity not specified in radar object, '
                       'provide this value explicitly in the function call.')
            raise ValueError(message)
        nyquist_vel = radar.instrument_parameters[
            'nyquist_velocity']['data'][0]
    return nyquist_vel


def _parse_gatefilter(gatefilter, radar, **kwargs):
    """ Parse the gatefilter, return a valid GateFilter object. """
    # parse the gatefilter parameter
    if gatefilter is None:  # create a moment based filter
        gatefilter = moment_based_gate_filter(radar, **kwargs)
    elif gatefilter is False:
        gatefilter = GateFilter(radar)
    else:
        gatefilter = gatefilter.copy()
    return gatefilter


def _parse_rays_wrap_around(rays_wrap_around, radar):
    """ Parse the rays_wrap_around parameter. """
    if rays_wrap_around is None:
        if radar.scan_type == 'ppi':
            rays_wrap_around = True
        else:
            rays_wrap_around = False
    return rays_wrap_around



"""
pyart.graph.coord_transform
===========================

Coordinate transform routines.

.. autosummary::
    :toctree: generated/

    radar_coords_to_cart_track_relative
    radar_coords_to_cart_earth_relative
    radar_coords_to_cart_aircraft_relative

"""

import numpy as np


def radar_coords_to_cart_track_relative(rng, rot, roll, drift, tilt, pitch):
    """
    Calculate track-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    rng : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    drift : array
        Drift angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    track-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    drift = np.radians(drift)           # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = rng * 1000.0                    # distances to gates in meters.

    x = r * (np.cos(rot + roll) * np.sin(drift) * np.cos(tilt) *
             np.sin(pitch) + np.cos(drift) * np.sin(rot + roll) *
             np.cos(tilt) - np.sin(drift) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(drift) * np.cos(tilt) *
             np.sin(pitch) + np.sin(drift) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(drift) * np.cos(pitch) * np.sin(tilt))
    z = (r * np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
         np.sin(pitch) * np.sin(tilt))
    return x, y, z


def radar_coords_to_cart_earth_relative(rng, rot, roll, heading, tilt, pitch):
    """
    Calculate earth-relative Cartesian coordinates from radar coordinates

    Parameters
    ----------
    rng : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    heading : array
        Heading (compass) angle of the radar in degrees clockwise from north.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    heading = np.radians(heading)       # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = rng * 1000.0                    # distances to gates in meters.

    x = r * (-1. * np.cos(rot + roll) * np.sin(heading) * np.cos(tilt) *
             np.sin(pitch) + np.cos(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.sin(heading) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(heading) * np.cos(tilt) *
             np.sin(pitch) - np.sin(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(heading) * np.cos(pitch) * np.sin(tilt))
    z = (r * np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
         np.sin(pitch) * np.sin(tilt))
    return x, y, z


def radar_coords_to_cart_aircraft_relative(rng, rot, tilt):
    """
    Calculate aircraft-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    rng : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.

    Returns
    -------
    X, Y, Z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    r = rng * 1000.0                    # distances to gates in meters.
    x = r * np.cos(tilt) * np.sin(rot)
    y = r * np.sin(tilt)
    z = r * np.cos(rot) * np.cos(tilt)
    return x, y, z



"""
pyart.io.mdv
============

Utilities for reading of MDV files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    MdvFile

.. autosummary::
    :toctree: generated/

    read_mdv

"""
# Code is adapted from Nitin Bharadwaj's Matlab code

import struct
import gzip
import zlib
import StringIO
import datetime

import numpy as np
from netCDF4 import date2num

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str
from .common import radar_coords_to_cart


def read_mdv(filename, field_names=None, additional_metadata=None,
             file_field_names=False, exclude_fields=None):
    """
    Read a MDV file.

    Parameters
    ----------
    filename : str
        Name of MDV file to read or file-like object pointing to the
        beginning of such a file.
    field_names : dict, optional
        Dictionary mapping MDV data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.

    Returns
    -------
    radar : Radar
        Radar object containing data from MDV file.

    Notes
    -----
    Currently this function can only read polar MDV files with fields
    compressed with gzip or zlib.

    """
    # create metadata retrieval object
    filemetadata = FileMetadata('mdv', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    mdvfile = MdvFile(filename)

    # value attributes
    naz = len(mdvfile.az_deg)
    nele = len(mdvfile.el_deg)
    scan_type = mdvfile.scan_type

    if scan_type not in ['ppi', 'rhi']:
        raise NotImplementedError('No support for scan_type %s.' % scan_type)

    # time
    time = filemetadata('time')
    units = make_time_unit_str(mdvfile.times['time_begin'])
    time['units'] = units
    time_start = date2num(mdvfile.times['time_begin'], units)
    time_end = date2num(mdvfile.times['time_end'], units)
    time['data'] = np.linspace(time_start, time_end, naz * nele)

    # range
    _range = filemetadata('range')
    _range['data'] = np.array(mdvfile.range_km * 1000.0, dtype='float32')
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    _range['meters_between_gates'] = (_range['data'][1] - _range['data'][0])

    # fields
    fields = {}
    for mdv_field in set(mdvfile.fields):
        field_name = filemetadata.get_field_name(mdv_field)
        if field_name is None:
            continue

        # grab data from MDV object, mask and reshape
        data = mdvfile.read_a_field(mdvfile.fields.index(mdv_field))
        data[np.where(np.isnan(data))] = get_fillvalue()
        data[np.where(data == 131072)] = get_fillvalue()
        data = np.ma.masked_equal(data, get_fillvalue())
        data.shape = (data.shape[0] * data.shape[1], data.shape[2])

        # create and store the field dictionary
        field_dic = filemetadata(field_name)
        field_dic['data'] = data
        field_dic['_FillValue'] = get_fillvalue()
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    for meta_key, mdv_key in MDV_METADATA_MAP.iteritems():
        metadata[meta_key] = mdvfile.master_header[mdv_key]

    # latitude
    latitude = filemetadata('latitude')
    latitude['data'] = np.array([mdvfile.radar_info['latitude_deg']],
                                dtype='float64')
    # longitude
    longitude = filemetadata('longitude')
    longitude['data'] = np.array([mdvfile.radar_info['longitude_deg']],
                                 dtype='float64')
    # altitude
    altitude = filemetadata('altitude')
    altitude['data'] = np.array([mdvfile.radar_info['altitude_km'] * 1000.0],
                                dtype='float64')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    len_time = len(time['data'])

    if mdvfile.scan_type == 'ppi':
        nsweeps = nele
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')
        sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])
        fixed_angle['data'] = np.array(mdvfile.el_deg, dtype='float32')
        sweep_start_ray_index['data'] = np.arange(0, len_time, naz,
                                                  dtype='int32')
        sweep_end_ray_index['data'] = np.arange(naz-1, len_time, naz,
                                                dtype='int32')

    elif mdvfile.scan_type == 'rhi':
        nsweeps = naz
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')
        sweep_mode['data'] = np.array(nsweeps * ['rhi'])
        fixed_angle['data'] = np.array(mdvfile.az_deg, dtype='float32')
        sweep_start_ray_index['data'] = np.arange(0, len_time, nele,
                                                  dtype='int32')
        sweep_end_ray_index['data'] = np.arange(nele - 1, len_time, nele,
                                                dtype='int32')

    # azimuth, elevation
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')

    if scan_type == 'ppi':
        azimuth['data'] = np.tile(mdvfile.az_deg, nele)
        elevation['data'] = np.array(mdvfile.el_deg).repeat(naz)

    elif scan_type == 'rhi':
        azimuth['data'] = np.array(mdvfile.az_deg).repeat(nele)
        elevation['data'] = np.tile(mdvfile.el_deg, naz)

    # instrument parameters
    # we will set 4 parameters in the instrument_parameters dict
    # prt, prt_mode, unambiguous_range, and nyquist_velocity

    # TODO prt mode: Need to fix this.. assumes dual if two prts
    if mdvfile.radar_info['prt2_s'] == 0.0:
        prt_mode_str = 'fixed'
    else:
        prt_mode_str = 'dual'

    prt_mode = filemetadata('prt_mode')
    prt = filemetadata('prt')
    unambiguous_range = filemetadata('unambiguous_range')
    nyquist_velocity = filemetadata('nyquist_velocity')
    beam_width_h = filemetadata('radar_beam_width_h')
    beam_width_v = filemetadata('radar_beam_width_v')

    prt_mode['data'] = np.array([prt_mode_str] * nsweeps)
    prt['data'] = np.array([mdvfile.radar_info['prt_s']] * nele * naz,
                           dtype='float32')

    urange_m = mdvfile.radar_info['unambig_range_km'] * 1000.0
    unambiguous_range['data'] = np.array([urange_m] * naz * nele,
                                         dtype='float32')

    uvel_mps = mdvfile.radar_info['unambig_vel_mps']
    nyquist_velocity['data'] = np.array([uvel_mps] * naz * nele,
                                        dtype='float32')
    beam_width_h['data'] = np.array(
        [mdvfile.radar_info['horiz_beam_width_deg']], dtype='float32')
    beam_width_v['data'] = np.array(
        [mdvfile.radar_info['vert_beam_width_deg']], dtype='float32')

    instrument_parameters = {'prt_mode': prt_mode, 'prt': prt,
                             'unambiguous_range': unambiguous_range,
                             'nyquist_velocity': nyquist_velocity,
                             'radar_beam_width_h': beam_width_h,
                             'radar_beam_width_v': beam_width_v}

    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


# mapping from MDV name space to CF-Radial name space
MDV_METADATA_MAP = {'instrument_name': 'data_set_source',
                    'source': 'data_set_info'}


# Information about the MDV file structure
MDV_CHUNK_INFO_LEN = 480
MDV_INFO_LEN = 512
MDV_LONG_FIELD_LEN = 64
MDV_MAX_PROJ_PARAMS = 8
MDV_MAX_VLEVELS = 122
MDV_NAME_LEN = 128
MDV_SHORT_FIELD_LEN = 16
MDV_TRANSFORM_LEN = 16
MDV_UNITS_LEN = 16
MDV_N_COORD_LABELS = 3
MDV_COORD_UNITS_LEN = 32

# (x,y) in degrees. Simple latitude-longitude grid.
# Also known as the Simple Cylindrical or Platte Carree projection.
PROJ_LATLON = 0
# (x,y) in km. Lambert Conformal Conic projection.
PROJ_LAMBERT_CONF = 3
# (x,y) in km. Polar Stereographic projection.
PROJ_POLAR_STEREO = 5
# Cartesian, (x,y) in km. This is a simple line-of-sight
# projection used for single radar sites. The formal name is
# Oblique Lambert Azimuthal projection.
PROJ_FLAT = 8
# radar data in native Plan Position Indicator (PPI) coordinates of
# range, azimuth angle and elevation angle. x is radial range (km),
# y is azimuth angle (deg), z is elev angle (deg).
PROJ_POLAR_RADAR = 9
# (x,y) in km. Oblique Stereographic projection.
PROJ_OBLIQUE_STEREO = 12
# radar data in native Range Height Indicator (RHI) coordinates.
# x is radial range (km), y is elev angle (deg), z is az angle (deg).
PROJ_RHI_RADAR = 13
#  ***************** COMPRESSION *******************
COMPRESSION_NONE = 0  # no compression
COMPRESSION_ZLIB = 3  # Lempel-Ziv
COMPRESSION_BZIP = 4  # bzip2
COMPRESSION_GZIP = 5  # Lempel-Ziv in gzip format

#  ***************** COMPRESSION CODE *******************
TA_NOT_COMPRESSED = 791621423
GZIP_COMPRESSED = 4160223223

#  ***************** TRANSFORM *******************
DATA_TRANSFORM_NONE = 0  # None
DATA_TRANSFORM_LOG = 1  # Natural log

#  ***************** BIT ENCODING *******************
ENCODING_INT8 = 1  # unsigned 8 bit integer
ENCODING_INT16 = 2  # unsigned 16 bit integer
ENCODING_FLOAT32 = 5  # 32 bit IEEE floating point

#  ***************** CHUNK HEADER and DATA *******************
CHUNK_DSRADAR_PARAMS = 3
CHUNK_DSRADAR_ELEVATIONS = 10
CHUNK_DSRADAR_CALIB = 7
DS_LABEL_LEN = 40
NCHAR_DS_RADAR_PARAMS = 2 * DS_LABEL_LEN
DS_RADAR_CALIB_NAME_LEN = 16
DS_RADAR_CALIB_MISSING = -9999.0


class MdvFile:
    """
    A file object for MDV data.

    A `MdvFile` object stores metadata and data from a MDV file.  Metadata is
    stored in dictionaries as attributes of the object, field data is
    stored as NumPy ndarrays as attributes with the field name. By default
    only metadata is read initially and field data must be read using the
    `read_a_field` or `read_all_fields` methods.  This behavior can be changed
    by setting the `read_fields` parameter to True.

    Parameters
    ----------
    filename : str or file-like
        Name of MDV file to read or file-like object pointing to the
        beginning of such a file.
    debug : bool
        True to print out debugging information, False to supress
    read_fields : bool
        True to read all field during initalization, False (default) only
        reads metadata.

    """

    def __init__(self, filename, debug=False, read_fields=False):
        """ initalize MdvFile from filename (str). """

        if debug:
            print "Opening ", filename
        if hasattr(filename, 'read'):
            self.fileptr = filename
        else:
            self.fileptr = open(filename, 'rb')

        if debug:
            print "Getting master header"
        self.master_header = self._get_master_header()

        if debug:
            print "getting field headers"
        nfields = self.master_header['nfields']
        self.field_headers = self._get_field_headers(nfields)

        if debug:
            print "getting vlevel headers"
        self.vlevel_headers = self._get_vlevel_headers(nfields)

        if debug:
            print "getting chunk headers"
        nchunks = self.master_header['nchunks']
        self.chunk_headers = self._get_chunk_headers(nchunks)

        if debug:
            print "Getting Chunk Data"
        radar_info, elevations, calib_info = self._get_chunks()
        if radar_info is not None:
            self.radar_info = radar_info
        if elevations is not None:
            self.elevations = elevations
        if calib_info is not None:
            self.calib_info = calib_info

        if debug:
            print "Calculating Radar coordinates"
        az_deg, range_km, el_deg = self._calc_geometry()
        self.az_deg = np.array(az_deg, dtype='float32')
        self.range_km = np.array(range_km, dtype='float32')
        self.el_deg = np.array(el_deg, dtype='float32')

        if debug:
            print "Making usable time objects"
        self.times = self._make_time_dict()

        if debug:
            print "Calculating cartesian coordinates"
        self.carts = self._make_carts_dict()

        if debug:
            print "indexing fields"
        self.fields = self._make_fields_list()

        if read_fields:
            if debug:
                print "Reading all fields"
            self.read_all_fields()
        return

    ##################
    # public methods #
    ##################

    def read_a_field(self, fnum, debug=False):
        """
        Read a field from the MDV file.

        Parameters
        ----------
        fnum : int
            Field number to read.
        debug : bool
            True to print debugging information, False to supress.

        Returns
        -------
        field_data : array
            Field data.  This data is also stored as a object attribute under
            the field name.

        See Also
        --------
        read_all_fields : Read all fields in the MDV file.

        """

        field_header = self.field_headers[fnum]
        # if the field has already been read, return it
        if hasattr(self, field_header['field_name']):
            if debug:
                print "Getting data from the object."
            return getattr(self, field_header['field_name'])

        # field has not yet been read, populate the object and return
        if debug:
            print "No data found in object, populating"

        nsweeps = self.master_header['nsweeps']
        nrays = self.master_header['nrays']
        ngates = self.master_header['ngates']

        # read the header
        field_data = np.zeros([nsweeps, nrays, ngates], dtype='float32')
        self.fileptr.seek(field_header['field_data_offset'])
        self._get_sweep_info(nsweeps)  # dict not used, but need to seek.

        for sw in xrange(nsweeps):
            if debug:
                print "doing sweep ", sw

            # get the compressed sweep data
            compr_info = self._get_compression_info()
            compr_data = self.fileptr.read(compr_info['nbytes_coded'])

            encoding_type = field_header['encoding_type']
            if encoding_type == ENCODING_INT8:
                fmt = '>%iB' % (ngates * nrays)
                np_form = '>B'
            elif encoding_type == ENCODING_INT16:
                fmt = '>%iH' % (ngates * nrays)
                np_form = '>H'
            elif encoding_type == ENCODING_FLOAT32:
                fmt = '>%if' % (ngates * nrays)
                np_form = '>f'
            else:
                raise ValueError('unknown encoding: ', encoding_type)

            # decompress the sweep data
            if compr_info['magic_cookie'] == 0xf7f7f7f7:
                cd_fobj = StringIO.StringIO(compr_data)
                gzip_file_handle = gzip.GzipFile(fileobj=cd_fobj)
                decompr_data = gzip_file_handle.read(struct.calcsize(fmt))
                gzip_file_handle.close()
            elif compr_info['magic_cookie'] == 0xf5f5f5f5:
                decompr_data = zlib.decompress(compr_data)
            else:
                raise NotImplementedError('unsupported compression mode')
                # With sample data it should be possible to write
                # decompressor for other modes, the compression magic
                # cookies for these modes are:
                # 0x2f2f2f2f : TA_NOT_COMPRESSED
                # 0xf8f8f8f8 : GZIP_NOT_COMPRSSED
                # 0xf3f3f3f3 : BZIP_COMPRESSED
                # 0xf4f4f4f4 : BZIP_NOT_COMPRESSED
                # 0xf6f6f6f6 : ZLIB_NOT_COMPRESSED

            # read the decompressed data, reshape and mask
            sw_data = np.fromstring(decompr_data, np_form).astype('float32')
            sw_data.shape = (nrays, ngates)
            mask = sw_data == field_header['bad_data_value']
            np.putmask(sw_data, mask, [np.NaN])

            # scale and offset the data, store in field_data
            scale = field_header['scale']
            bias = field_header['bias']
            field_data[sw, :, :] = sw_data * scale + bias

        # store data as object attribute and return
        setattr(self, field_header['field_name'], field_data)
        return field_data

    def read_all_fields(self):
        """ Read all fields, storing data to field name attributes. """
        for i in xrange(self.master_header['nfields']):
            self.read_a_field(i)

    def close(self):
        """ Close the MDV file. """
        self.fileptr.close()

    ###################
    # private methods #
    ###################

    # get_ methods for reading headers

    def _get_master_header(self):
        """ Read the MDV master header, return a dict. """
        fmt = '>28i 8i i 5i 6f 3f 12f 512c 128c 128c i'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["record_len1"] = l[0]             # 28i: 1
        d["struct_id"] = l[1]               # 28i: 2
        d["revision_number"] = l[2]         # 28i: 3
        d["time_gen"] = l[3]                # 28i: 4
        d["user_time"] = l[4]               # 28i: 5
        d["time_begin"] = l[5]              # 28i: 6
        d["time_end"] = l[6]                # 28i: 7
        d["time_centroid"] = l[7]           # 28i: 8
        d["time_expire"] = l[8]             # 28i: 9
        d["num_data_times"] = l[9]          # 28i: 10
        d["index_number"] = l[10]           # 28i: 11
        d["data_dimension"] = l[11]         # 28i: 12
        d["data_collection_type"] = l[12]   # 28i: 13
        d["user_data"] = l[13]              # 28i: 14
        d["native_vlevel_type"] = l[14]     # 28i: 15
        d["vlevel_type"] = l[15]            # 28i: 16
        d["vlevel_included"] = l[16]        # 28i: 17
        d["grid_orientation"] = l[17]       # 28i: 18
        d["data_ordering"] = l[18]          # 28i: 19
        d["nfields"] = l[19]                # 28i: 20
        d["ngates"] = l[20]                 # 28i: 21
        d["nrays"] = l[21]                  # 28i: 22
        d["nsweeps"] = l[22]                # 28i: 23
        d["nchunks"] = l[23]                # 28i: 24
        d["field_hdr_offset"] = l[24]       # 28i: 25
        d["vlevel_hdr_offset"] = l[25]      # 28i: 26
        d["chunk_hdr_offset"] = l[26]       # 28i: 27
        d["field_grids_differ"] = l[27]     # 28i: 28
        d["user_data_si328"] = l[28:36]     # 8i
        d["time_written"] = l[36]           # i
        d["unused_si325"] = l[37:42]        # 5i
        d["user_data_fl326"] = l[42:48]     # 6f
        d["sensor_lon"] = l[48]             # 3f : 1
        d["sensor_lat"] = l[49]             # 3f : 2
        d["sensor_alt"] = l[50]             # 3f : 3
        d["unused_fl3212"] = l[51:63]       # 12f
        d["data_set_info"] = ''.join(l[63:575]).strip('\x00')       # 512c
        d["data_set_name"] = ''.join(l[575:703]).strip('\x00')      # 128c
        d["data_set_source"] = ''.join(l[703:831]).strip('\x00')    # 128c
        d["record_len2"] = l[831]           # i
        return d

    def _get_field_headers(self, nfields):
        """ Read nfields field headers, return a list of dicts. """
        return [self._get_field_header() for i in range(nfields)]

    def _get_field_header(self):
        """ Read a single field header, return a dict. """

        fmt = '>17i 10i 9i 4i f f 8f 12f 4f 5f 64c 16c 16c 16c 16c i'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["record_len1"] = l[0]             # 17i: 1
        d["struct_id"] = l[1]               # 17i: 2
        d["field_code"] = l[2]              # 17i: 3
        d["user_time1"] = l[3]              # 17i: 4
        d["forecast_delta"] = l[4]          # 17i: 5
        d["user_time2"] = l[5]              # 17i: 6
        d["user_time3"] = l[6]              # 17i: 7
        d["forecast_time"] = l[7]           # 17i: 8
        d["user_time4"] = l[8]              # 17i: 9
        d["ngates"] = l[9]                  # 17i: 10
        d["nrays"] = l[10]                  # 17i: 11
        d["nsweeps"] = l[11]                # 17i: 12
        d["proj_type"] = l[12]              # 17i: 13
        d["encoding_type"] = l[13]          # 17i: 14
        d["data_element_nbytes"] = l[14]    # 17i: 15
        d["field_data_offset"] = l[15]      # 17i: 16
        d["volume_size"] = l[16]            # 17i: 17
        d["user_data_si32"] = l[17:27]      # 10i
        d["compression_type"] = l[27]       # 9i: 1
        d["transform_type"] = l[28]         # 9i: 2
        d["scaling_type"] = l[29]           # 9i: 3
        d["native_vlevel_type"] = l[30]     # 9i: 4
        d["vlevel_type"] = l[31]            # 9i: 5
        d["dz_constant"] = l[32]            # 9i: 6
        d["data_dimension"] = l[33]         # 9i: 7
        d["zoom_clipped"] = l[34]           # 9i: 8
        d["zoom_no_overlap"] = l[35]        # 9i: 9
        d["unused_si32"] = l[36:40]         # 4i
        d["proj_origin_lat"] = l[40]        # f
        d["proj_origin_lon"] = l[41]        # f
        d["proj_param"] = l[42:50]          # 8f
        d["vert_reference"] = l[50]         # 12f: 1
        d["grid_dx"] = l[51]                # 12f: 2
        d["grid_dy"] = l[52]                # 12f: 3
        d["grid_dz"] = l[53]                # 12f: 4
        d["grid_minx"] = l[54]              # 12f: 5
        d["grid_miny"] = l[55]              # 12f: 6
        d["grid_minz"] = l[56]              # 12f: 7
        d["scale"] = l[57]                  # 12f: 8
        d["bias"] = l[58]                   # 12f: 9
        d["bad_data_value"] = l[59]         # 12f: 10
        d["missing_data_value"] = l[60]     # 12f: 11
        d["proj_rotation"] = l[61]          # 12f: 12
        d["user_data_fl32"] = l[62:66]      # 4f
        d["min_value"] = l[66]              # 5f: 1
        d["max_value"] = l[67]              # 5f: 2
        d["min_value_orig_vol"] = l[68]     # 5f: 3
        d["max_value_orig_vol"] = l[69]     # 5f: 4
        d["unused_fl32"] = l[70]            # 5f: 5
        d["field_name_long"] = ''.join(l[71:135]).strip('\x00')     # 64c
        d["field_name"] = ''.join(l[135:151]).strip('\x00')         # 16c
        d["units"] = ''.join(l[151:167]).strip('\x00')              # 16c
        d["transform"] = ''.join(l[167:183]).strip('\x00')          # 16c
        d["unused_char"] = ''.join(l[183:199]).strip('\x00')        # 16c
        d["record_len2"] = l[199]           # i
        return d

    def _get_vlevel_headers(self, nfields):
        """ Read nfields vlevel headers, return a list of dicts. """
        return [self._get_vlevel_header() for i in range(nfields)]

    def _get_vlevel_header(self):
        """ Read a single vlevel header, return a dict. """
        fmt = '>i i 122i 4i 122f 5f i'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["record_len1"] = l[0]         # i
        d["struct_id"] = l[1]           # i
        d["type"] = l[2:124]            # 122i
        d["unused_si32"] = l[124:128]   # 4i
        d["level"] = l[128:250]         # 122f
        d["unused_fl32"] = l[250:255]   # 5f
        d["record_len2"] = l[255]       # i
        return d

    def _get_chunk_headers(self, nchunks):
        """ Get nchunk chunk headers, return a list of dicts. """
        return [self._get_chunk_header() for i in range(nchunks)]

    def _get_chunk_header(self):
        """ Get a single chunk header, return a dict. """
        fmt = '>5i 2i 480c i'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["record_len1"] = l[0]         # 5i: 1
        d["struct_id"] = l[1]           # 5i: 2
        d["chunk_id"] = l[2]            # 5i: 3
        d["chunk_data_offset"] = l[3]   # 5i: 4
        d["size"] = l[4]                # 5i: 5
        d["unused_si32"] = l[5:7]       # 2i
        d["info"] = ''.join(l[7:487]).strip('\x00')     # 480c
        d["record_len2"] = l[487]       # i
        return d

    def _get_chunks(self, debug=False):
        """ Get data in chunks, return radar_info, elevations, calib_info. """
        radar_info, elevations, calib_info = None, None, None
        for curr_chunk_header in self.chunk_headers:

            chunk_id = curr_chunk_header['chunk_id']
            self.fileptr.seek(curr_chunk_header['chunk_data_offset'])

            if chunk_id == CHUNK_DSRADAR_PARAMS:
                if debug:
                    print 'Getting radar info'
                radar_info = self._get_radar_info()

            elif chunk_id == CHUNK_DSRADAR_CALIB:
                if debug:
                    print 'getting elevations'
                elevations = self._get_elevs(curr_chunk_header['size'])

            elif chunk_id == CHUNK_DSRADAR_ELEVATIONS:
                if debug:
                    print 'getting cal'
                calib_info = self._get_calib()

        return radar_info, elevations, calib_info

    def _get_radar_info(self):
        """ Get the radar information, return dict. """
        fmt = '>12i 2i 22f 4f 40c 40c'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["radar_id"] = l[0]                # 12i: 1
        d["radar_type"] = l[1]              # 12i: 2
        d["nfields"] = l[2]                 # 12i: 3
        d["ngates"] = l[3]                  # 12i: 4
        d["samples_per_beam"] = l[4]        # 12i: 5
        d["scan_type"] = l[5]               # 12i: 6
        d["scan_mode"] = l[6]               # 12i: 7
        d["nfields_current"] = l[7]         # 12i: 8
        d["field_flag"] = l[8]              # 12i: 9
        d["polarization"] = l[9]            # 12i: 10
        d["follow_mode"] = l[10]            # 12i: 11
        d["prf_mode"] = l[11]               # 12i: 12
        d["spare_ints"] = l[12:14]          # 2i
        d["radar_constant"] = l[14]         # 22f: 1
        d["altitude_km"] = l[15]            # 22f: 2
        d["latitude_deg"] = l[16]           # 22f: 3
        d["longitude_deg"] = l[17]          # 22f: 4
        d["gate_spacing_km"] = l[18]        # 22f: 5
        d["start_range_km"] = l[19]         # 22f: 6
        d["horiz_beam_width_deg"] = l[20]   # 22f: 7
        d["vert_beam_width_deg"] = l[21]    # 22f: 8
        d["pulse_width_us"] = l[22]         # 22f: 9
        d["prf_hz"] = l[23]                 # 22f: 10
        d["wavelength_cm"] = l[24]          # 22f: 11
        d["xmit_peak_pwr_watts"] = l[25]    # 22f: 12
        d["receiver_mds_dbm"] = l[26]       # 22f: 13
        d["receiver_gain_db"] = l[27]       # 22f: 14
        d["antenna_gain_db"] = l[28]        # 22f: 15
        d["system_gain_db"] = l[29]         # 22f: 16
        d["unambig_vel_mps"] = l[30]        # 22f: 17
        d["unambig_range_km"] = l[31]       # 22f: 18
        d["measXmitPowerDbmH_dbm"] = l[32]  # 22f: 19
        d["measXmitPowerDbmV_dbm"] = l[33]  # 22f: 20
        d["prt_s"] = l[34]                  # 22f: 21
        d["prt2_s"] = l[35]                 # 22f: 22
        d["spare_floats"] = l[36:40]        # 4f
        d["radar_name"] = ''.join(l[40:80]).strip('\x00')       # 40c
        d["scan_type_name"] = ''.join(l[80:120]).strip('\x00')   # 40c
        return d

    def _get_elevs(self, nbytes):
        """ Return an array of elevation read from current file position. """
        SIZE_FLOAT = 4.0
        nelevations = np.floor(nbytes / SIZE_FLOAT)
        fmt = '%df' % (nelevations)
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        return np.array(l)

    def _get_calib(self):
        """ Get the calibration information, return a dict. """
        fmt = '>16c 6i 51f 14f'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d["radar_name"] = ''.join(l[0:16]).strip('\x00')    # c16
        d["year"] = l[16]                           # 6i: 1
        d["month"] = l[17]                          # 6i: 2
        d["day"] = l[18]                            # 6i: 3
        d["hour"] = l[19]                           # 6i: 4
        d["minute"] = l[20]                         # 6i: 5
        d["second"] = l[21]                         # 6i: 6
        d["wavelength_cm"] = l[22]                  # 51f: 1
        d["beamwidth_h_deg"] = l[23]                # 51f: 2
        d["beamwidth_v_deg"] = l[24]                # 51f: 3
        d["antenna_gain_h_db"] = l[25]              # 51f: 4
        d["antenna_gain_v_db"] = l[26]              # 51f: 5
        d["pulse_width_us"] = l[27]                 # 51f: 6
        d["xmit_power_h_dbm"] = l[28]               # 51f: 7
        d["xmit_power_v_dbm"] = l[29]               # 51f: 8
        d["twoway_waveguide_loss_h_db"] = l[30]     # 51f: 9
        d["twoway_waveguide_loss_v_db"] = l[31]     # 51f: 10
        d["twoway_radome_loss_h_db"] = l[32]        # 51f: 11
        d["twoway_radome_loss_v_db"] = l[33]        # 51f: 12
        d["filter_loss_db"] = l[34]                 # 51f: 13
        d["radar_constant_h_db"] = l[35]            # 51f: 14
        d["radar_constant_v_db"] = l[36]            # 51f: 15
        d["noise_h_co_dbm"] = l[37]                 # 51f: 16
        d["noise_h_cx_dbm"] = l[38]                 # 51f: 17
        d["noise_v_co_dbm"] = l[39]                 # 51f: 18
        d["noise_v_cx_dbm"] = l[40]                 # 51f: 19
        d["rx_gain_h_co_dbm"] = l[41]               # 51f: 20
        d["rx_gain_h_cx_dbm"] = l[42]               # 51f: 21
        d["rx_gain_v_co_dbm"] = l[43]               # 51f: 22
        d["rx_gain_v_cx_dbm"] = l[44]               # 51f: 23
        d["zh1km_co_dbz"] = l[45]                   # 51f: 24
        d["zh1km_cx_dbz"] = l[46]                   # 51f: 25
        d["zv1km_co_dbz"] = l[47]                   # 51f: 26
        d["zv1km_cx_dbz"] = l[48]                   # 51f: 27
        d["sun_h_co_dbm"] = l[49]                   # 51f: 28
        d["sun_h_cx_dbm"] = l[50]                   # 51f: 29
        d["sun_v_co_dbm"] = l[51]                   # 51f: 30
        d["sun_v_cx_dbm"] = l[52]                   # 51f: 31
        d["noise_source_h_dbm"] = l[53]             # 51f: 32
        d["noise_source_v_dbm"] = l[54]             # 51f: 33
        d["power_meas_loss_h_db"] = l[55]           # 51f: 34
        d["power_meas_loss_v_db"] = l[56]           # 51f: 35
        d["coupler_fwd_loss_h_db"] = l[57]          # 51f: 36
        d["coupler_fwd_loss_v_db"] = l[58]          # 51f: 37
        d["zdr_bias_db"] = l[59]                    # 51f: 38
        d["ldr_h_bias_db"] = l[60]                  # 51f: 39
        d["ldr_v_bias_db"] = l[61]                  # 51f: 40
        d["system_phidp_deg"] = l[62]               # 51f: 41
        d["test_pulse_h_dbm"] = l[63]               # 51f: 42
        d["test_pulse_v_dbm"] = l[64]               # 51f: 43
        d["rx_slope_h_co_db"] = l[65]               # 51f: 44
        d["rx_slope_h_cx_db"] = l[66]               # 51f: 45
        d["rx_slope_v_co_db"] = l[67]               # 51f: 46
        d["rx_slope_v_cx_db"] = l[68]               # 51f: 47
        d["I0_h_co_dbm"] = l[69]                    # 51f: 48
        d["I0_h_cx_dbm"] = l[70]                    # 51f: 49
        d["I0_v_co_dbm"] = l[71]                    # 51f: 50
        d["I0_v_cx_dbm"] = l[72]                    # 51f: 51
        d["spare"] = l[73:87]                       # 14f
        return d

    def _get_compression_info(self):
        """ Get compression infomation, return a dict. """
        # the file pointer must be set at the correct location prior to call
        fmt = '>I I I I 2I'
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d['magic_cookie'] = l[0]
        d['nbytes_uncompressed'] = l[1]
        d['nbytes_compressed'] = l[2]
        d['nbytes_coded'] = l[3]
        d['spare'] = l[4:6]
        return d

    def _get_sweep_info(self, nsweeps):
        """ Get sweep information, return a dict. """
        # the file pointer must be set at the correct location prior to call
        fmt = '%iI %iI' % (nsweeps, nsweeps)
        l = struct.unpack(fmt, self.fileptr.read(struct.calcsize(fmt)))
        d = {}
        d['vlevel_offsets'] = l[:nsweeps]
        d['vlevel_nbytes'] = l[nsweeps:2*nsweeps]
        return d

    # misc. methods

    def _calc_geometry(self):
        """ Calculate geometry, return az_deg, range_km, el_deg. """
        nsweeps = self.master_header['nsweeps']
        nrays = self.master_header['nrays']
        ngates = self.master_header['ngates']
        grid_minx = self.field_headers[0]['grid_minx']
        grid_miny = self.field_headers[0]['grid_miny']
        grid_dx = self.field_headers[0]['grid_dx']
        grid_dy = self.field_headers[0]['grid_dy']

        range_km = grid_minx + np.arange(ngates) * grid_dx

        if self.field_headers[0]['proj_type'] == PROJ_RHI_RADAR:
            self.scan_type = 'rhi'
            el_deg = grid_miny + np.arange(nrays) * grid_dy
            az_deg = self.vlevel_headers[0]['level'][0:nsweeps]

        if self.field_headers[0]['proj_type'] == PROJ_POLAR_RADAR:
            self.scan_type = 'ppi'
            az_deg = grid_miny + np.arange(nrays) * grid_dy
            el_deg = self.vlevel_headers[0]['level'][0:nsweeps]

        return az_deg, range_km, el_deg

    def _make_time_dict(self):
        """ Return a time dictionary. """
        t_base = datetime.datetime(1970, 1, 1, 00, 00)
        tb = datetime.timedelta(seconds=self.master_header['time_begin'])
        te = datetime.timedelta(seconds=self.master_header['time_end'])
        tc = datetime.timedelta(seconds=self.master_header['time_centroid'])
        return {'time_begin': t_base + tb, 'time_end': t_base + te,
                'time_centroid': t_base + tc}

    def _make_carts_dict(self):
        """ Return a carts dictionary, distances in meters. """

        # simple calculation involving 4/3 earth radius
        nsweeps = self.master_header['nsweeps']
        nrays = self.master_header['nrays']
        ngates = self.master_header['ngates']
        xx = np.empty([nsweeps, nrays, ngates], dtype=np.float32)
        yy = np.empty([nsweeps, nrays, ngates], dtype=np.float32)
        zz = np.empty([nsweeps, nrays, ngates], dtype=np.float32)

        if self.scan_type == 'rhi':
            rg, ele = np.meshgrid(self.range_km, self.el_deg)
            rg = np.array(rg, dtype=np.float64)
            ele = np.array(ele, dtype=np.float64)
            for aznum in xrange(nsweeps):
                azg = np.ones(rg.shape, dtype=np.float64) * self.az_deg[aznum]
                x, y, z = radar_coords_to_cart(rg, azg, ele)
                zz[aznum, :, :] = z
                xx[aznum, :, :] = x
                yy[aznum, :, :] = y

        elif self.scan_type == 'ppi':
            rg, azg = np.meshgrid(self.range_km, self.az_deg)
            rg = np.array(rg, dtype=np.float64)
            azg = np.array(azg, dtype=np.float64)
            for elnum in xrange(nsweeps):
                ele = np.ones(rg.shape, dtype=np.float64) * self.el_deg[elnum]
                x, y, z = radar_coords_to_cart(rg, azg, ele)
                zz[elnum, :, :] = z
                xx[elnum, :, :] = x
                yy[elnum, :, :] = y

        return {'x': xx, 'y': yy, 'z': zz}

    def _make_fields_list(self):
        """ Return a list of fields. """
        fh = self.field_headers
        return [fh[i]['field_name'] for i in range(len(fh))]



"""
==================================
Input and output (:mod:`pyart.io`)
==================================

.. currentmodule:: pyart.io

Py-ART has modules, classes and functions which are able to read data
from and write data to a number of file formats.

.. autosummary::
    :toctree: generated/

    read
    read_rsl
    read_mdv
    read_sigmet
    read_cfradial
    read_chl
    read_nexrad_archive
    read_nexrad_cdm
    write_cfradial
    read_grid
    write_grid

"""

try:
    from .rsl import read_rsl
    _RSL_AVAILABLE = True
except ImportError:
    _RSL_AVAILABLE = False
from .mdv import read_mdv
from .sigmet import read_sigmet
from .chl import read_chl
from .cfradial import read_cfradial, write_cfradial
from .nexrad_archive import read_nexrad_archive
from .nexrad_cdm import read_nexrad_cdm
from .grid_io import read_grid, write_grid
from .auto_read import read

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.graph.common
==================

Common graphing routines.

.. autosummary::
    :toctree: generated/

    dms_to_d
    radar_coords_to_cart
    corner_to_point
    ax_radius
    sweep_coords_to_cart
    _interpolate_range_edges
    _interpolate_elevation_edges
    _interpolate_azimuth_edges

"""

import numpy as np

PI = 3.141592653589793

from ..io.common import dms_to_d, radar_coords_to_cart


def corner_to_point(corner, point):
    """
    Return the x, y distances in meters from a corner to a point.

    Assumes a spherical earth model.

    Parameters
    ----------
    corner : (float, float)
        Latitude and longitude in degrees of the corner.
    point : (float, float)
        Latitude and longitude in degrees of the point.

    Returns
    -------
    x, y : floats
        Distances from the corner to the point in meters.

    """
    Re = 6371.0 * 1000.0
    Rc = ax_radius(point[0], units='degrees')
    #print Rc/Re
    y = ((point[0] - corner[0]) / 360.0) * PI * 2.0 * Re
    x = ((point[1] - corner[1]) / 360.0) * PI * 2.0 * Rc
    return x, y


def ax_radius(lat, units='radians'):
    """
    Return the radius of a constant latitude circle for a given latitude.

    Parameters
    ----------
    lat : float
        Latitude at which to calculate constant latitude circle (parallel)
        radius.
    units : 'radians' or 'degrees'
        Units of lat, either 'radians' or 'degrees'.

    Returns
    -------
    R : float
        Radius in meters of a constant latitude circle (parallel).

    """
    Re = 6371.0 * 1000.0
    if units == 'degrees':
        const = PI / 180.0
    else:
        const = 1.0
    R = Re * np.sin(PI / 2.0 - abs(lat * const))
    return R


def sweep_coords_to_cart(ranges, azimuths, elevations, edges=False):
    """
    Calculate Cartesian coordinate for the gates in a single radar sweep.

    Calculates the Cartesian coordinates for the gate centers or edges for
    all gates in a single sweep assuming a standard atmosphere (4/3 Earth's
    radius model). See :py:func:`pyart.io.common.radar_coords_to_cart` for
    details.

    Parameters
    ----------
    ranges : array, 1D.
        Distances to the center of the radar gates (bins) in meters.
    azimuths : array, 1D.
        Azimuth angles of the sweep in degrees.
    elevations : array, 1D.
        Elevation angles of the sweep in degrees.
    edges : bool, optional
        True to calculate the coordinates of the gate edges by interpolating
        between gates and extrapolating at the boundaries.  False to
        calculate the gate centers.

    Returns
    -------
    x, y, z : array, 2D
        Cartesian coordinates in meters from the center of the radar to the
        gate centers or edges.

    """
    if edges:
        if len(ranges) != 1:
            ranges = _interpolate_range_edges(ranges)
        if len(elevations) != 1:
            elevations = _interpolate_elevation_edges(elevations)
        if len(azimuths) != 1:
            azimuths = _interpolate_azimuth_edges(azimuths)
    rg, azg = np.meshgrid(ranges, azimuths)
    rg, eleg = np.meshgrid(ranges, elevations)
    return radar_coords_to_cart(rg / 1000., azg, eleg)


def _interpolate_range_edges(ranges):
    """ Interpolate the edges of the range gates from their centers. """
    edges = np.empty((ranges.shape[0] + 1, ), dtype=ranges.dtype)
    edges[1:-1] = (ranges[:-1] + ranges[1:]) / 2.
    edges[0] = ranges[0] - (ranges[1] - ranges[0]) / 2.
    edges[-1] = ranges[-1] - (ranges[-2] - ranges[-1]) / 2.
    edges[edges < 0] = 0    # do not allow range to become negative
    return edges


def _interpolate_elevation_edges(elevations):
    """ Interpolate the edges of the elevation angles from their centers. """
    edges = np.empty((elevations.shape[0]+1, ), dtype=elevations.dtype)
    edges[1:-1] = (elevations[:-1] + elevations[1:]) / 2.
    edges[0] = elevations[0] - (elevations[1] - elevations[0]) / 2.
    edges[-1] = elevations[-1] - (elevations[-2] - elevations[-1]) / 2.
    edges[edges > 180] = 180.   # prevent angles from going below horizon
    edges[edges < 0] = 0.
    return edges


def _interpolate_azimuth_edges(azimuths):
    """ Interpolate the edges of the azimuth angles from their centers. """
    edges = np.empty((azimuths.shape[0]+1, ), dtype=azimuths.dtype)
    # perform interpolation and extrapolation in complex plane to
    # account for periodic nature of azimuth angle.
    az = np.exp(1.j*np.deg2rad(azimuths))
    edges[1:-1] = np.angle((az[1:] + az[:-1]) / 2., deg=True)
    edges[0] = np.angle(az[0] - (az[1] - az[0]) / 2., deg=True)
    edges[-1] = np.angle(az[-1] - (az[-2] - az[-1]) / 2., deg=True)
    edges[edges < 0] += 360     # range from [-180, 180] to [0, 360]
    return edges



""" Unit Tests for Py-ART's graph/common.py module. """

import pyart


def test_corner_to_point():
    corner = (36.5, -97.5)
    point = (36.4, -97.6)
    x, y = pyart.graph.common.corner_to_point(corner, point)
    assert round(x) == -8950.
    assert round(y) == -11119.0


def test_ax_radius_degrees():
    R = pyart.graph.common.ax_radius(36.5, units='degrees')
    # answer found from: 6,371,000 * cos(36.5 / 180. * pi)
    assert round(R) == 5121372.


def test_ax_radius_radians():
    R = pyart.graph.common.ax_radius(0.637045177, units='radians')
    assert round(R) == 5121372.



"""
pyart.io.cfradial
=================

Utilities for reading CF/Radial files.

.. autosummary::
    :toctree: generated/

    read_cfradial
    write_cfradial
    _find_all_meta_group_vars
    _ncvar_to_dict
    _unpack_variable_gate_field_dic
    _create_ncvar

"""

import getpass
import datetime
import platform

import numpy as np
import netCDF4

from ..config import FileMetadata
from .common import stringarray_to_chararray
from ..core.radar import Radar


# Variables and dimensions in the instrument_parameter convention and
# radar_parameters sub-convention that will be read from and written to
# CfRadial files using Py-ART.
# The meta_group attribute cannot be used to identify these parameters as
# it is often set incorrectly.
_INSTRUMENT_PARAMS_DIMS = {
    # instrument_parameters sub-convention
    'frequency': ('frequency'),
    'follow_mode': ('sweep', 'string_length'),
    'pulse_width': ('time', ),
    'prt_mode': ('sweep', 'string_length'),
    'prt': ('time', ),
    'prt_ratio': ('time', ),
    'polarization_mode': ('sweep', 'string_length'),
    'nyquist_velocity': ('time', ),
    'unambiguous_range': ('time', ),
    'n_samples': ('time', ),
    'sampling_ratio': ('time', ),
    # radar_parameters sub-convention
    'radar_antenna_gain_h': (),
    'radar_antenna_gain_v': (),
    'radar_beam_width_h': (),
    'radar_beam_width_v': (),
    'radar_reciever_bandwidth': (),
    'radar_measured_transmit_power_h': ('time', ),
    'radar_measured_transmit_power_v': ('time', ),
    'radar_rx_bandwidth': (),           # non-standard
    'measured_transmit_power_v': ('time', ),    # non-standard
    'measured_transmit_power_h': ('time', ),    # non-standard
}


def read_cfradial(filename, field_names=None, additional_metadata=None,
                  file_field_names=False, exclude_fields=None):
    """
    Read a Cfradial netCDF file.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.

    Returns
    -------
    radar : Radar
        Radar object.

    Notes
    -----
    This function has not been tested on "stream" Cfradial files.

    """
    # create metadata retrieval object
    filemetadata = FileMetadata('cfradial', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])
    if 'n_gates_vary' in metadata:
        metadata['n_gates_vary'] = 'false'  # corrected below

    # 4.2 Dimensions (do nothing)

    # 4.3 Global variable -> move to metadata dictionary
    if 'volume_number' in ncvars:
        metadata['volume_number'] = int(ncvars['volume_number'][:])
    else:
        metadata['volume_number'] = 0

    global_vars = {'platform_type': 'fixed', 'instrument_type': 'radar',
                   'primary_axis': 'axis_z'}
    # ignore time_* global variables, these are calculated from the time
    # variable when the file is written.
    for var, default_value in global_vars.iteritems():
        if var in ncvars:
            metadata[var] = str(netCDF4.chartostring(ncvars[var][:]))
        else:
            metadata[var] = default_value

    # 4.4 coordinate variables -> create attribute dictionaries
    time = _ncvar_to_dict(ncvars['time'])
    _range = _ncvar_to_dict(ncvars['range'])

    # 4.5 Ray dimension variables

    # 4.6 Location variables -> create attribute dictionaries
    latitude = _ncvar_to_dict(ncvars['latitude'])
    longitude = _ncvar_to_dict(ncvars['longitude'])
    altitude = _ncvar_to_dict(ncvars['altitude'])
    if 'altitude_agl' in ncvars:
        altitude_agl = _ncvar_to_dict(ncvars['altitude_agl'])
    else:
        altitude_agl = None

    # 4.7 Sweep variables -> create atrribute dictionaries
    sweep_number = _ncvar_to_dict(ncvars['sweep_number'])
    sweep_mode = _ncvar_to_dict(ncvars['sweep_mode'])
    fixed_angle = _ncvar_to_dict(ncvars['fixed_angle'])
    sweep_start_ray_index = _ncvar_to_dict(ncvars['sweep_start_ray_index'])
    sweep_end_ray_index = _ncvar_to_dict(ncvars['sweep_end_ray_index'])
    if 'target_scan_rate' in ncvars:
        target_scan_rate = _ncvar_to_dict(ncvars['target_scan_rate'])
    else:
        target_scan_rate = None
    if 'rays_are_indexed' in ncvars:
        rays_are_indexed = _ncvar_to_dict(ncvars['rays_are_indexed'])
    else:
        rays_are_indexed = None
    if 'ray_angle_res' in ncvars:
        ray_angle_res = _ncvar_to_dict(ncvars['ray_angle_res'])
    else:
        ray_angle_res = None

    # first sweep mode determines scan_type
    mode = str(netCDF4.chartostring(sweep_mode['data'][0]))

    # options specified in the CF/Radial standard
    if mode == 'rhi':
        scan_type = 'rhi'
    elif mode == 'vertical_pointing':
        scan_type = 'vpt'
    elif mode == 'azimuth_surveillance':
        scan_type = 'ppi'
    elif mode == 'elevation_surveillance':
        scan_type = 'rhi'
    elif mode == 'manual_ppi':
        scan_type = 'ppi'
    elif mode == 'manual_rhi':
        scan_type = 'rhi'

    # fallback types
    elif 'sur' in mode:
        scan_type = 'ppi'
    elif 'sec' in mode:
        scan_type = 'sector'
    elif 'rhi' in mode:
        scan_type = 'rhi'
    else:
        scan_type = 'other'

    # 4.8 Sensor pointing variables -> create attribute dictionaries
    azimuth = _ncvar_to_dict(ncvars['azimuth'])
    elevation = _ncvar_to_dict(ncvars['elevation'])
    if 'scan_rate' in ncvars:
        scan_rate = _ncvar_to_dict(ncvars['scan_rate'])
    else:
        scan_rate = None

    if 'antenna_transition' in ncvars:
        antenna_transition = _ncvar_to_dict(ncvars['antenna_transition'])
    else:
        antenna_transition = None

    # 4.9 Moving platform geo-reference variables
    # Aircraft specific varaibles
    if 'rotation' in ncvars:
        rotation = _ncvar_to_dict(ncvars['rotation'])
    else:
        rotation = None

    if 'tilt' in ncvars:
        tilt = _ncvar_to_dict(ncvars['tilt'])
    else:
        tilt = None

    if 'roll' in ncvars:
        roll = _ncvar_to_dict(ncvars['roll'])
    else:
        roll = None

    if 'drift' in ncvars:
        drift = _ncvar_to_dict(ncvars['drift'])
    else:
        drift = None

    if 'heading' in ncvars:
        heading = _ncvar_to_dict(ncvars['heading'])
    else:
        heading = None

    if 'pitch' in ncvars:
        pitch = _ncvar_to_dict(ncvars['pitch'])
    else:
        pitch = None

    if 'georefs_applied' in ncvars:
        georefs_applied = _ncvar_to_dict(ncvars['georefs_applied'])
    else:
        georefs_applied = None

    # 4.10 Moments field data variables -> field attribute dictionary
    if 'ray_n_gates' in ncvars:
        # all variables with dimensions of n_points are fields.
        keys = [k for k, v in ncvars.iteritems()
                if v.dimensions == ('n_points', )]
    else:
        # all variables with dimensions of 'time', 'range' are fields
        keys = [k for k, v in ncvars.iteritems()
                if v.dimensions == ('time', 'range')]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            field_name = key
        fields[field_name] = _ncvar_to_dict(ncvars[key])

    if 'ray_n_gates' in ncvars:
        shape = (len(ncvars['time']), len(ncvars['range']))
        ray_n_gates = ncvars['ray_n_gates'][:]
        ray_start_index = ncvars['ray_start_index'][:]
        for dic in fields.values():
            _unpack_variable_gate_field_dic(
                dic, shape, ray_n_gates, ray_start_index)

    # 4.5 instrument_parameters sub-convention -> instrument_parameters dict
    # 4.6 radar_parameters sub-convention -> instrument_parameters dict
    keys = [k for k in _INSTRUMENT_PARAMS_DIMS.keys() if k in ncvars]
    instrument_parameters = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if instrument_parameters == {}:  # if no parameters set to None
        instrument_parameters = None

    # 4.7 lidar_parameters sub-convention -> skip

    # 4.8 radar_calibration sub-convention -> radar_calibration
    keys = _find_all_meta_group_vars(ncvars, 'radar_calibration')
    radar_calibration = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if radar_calibration == {}:
        radar_calibration = None

    ncobj.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        radar_calibration=radar_calibration,
        altitude_agl=altitude_agl,
        scan_rate=scan_rate,
        antenna_transition=antenna_transition,
        target_scan_rate=target_scan_rate,
        rays_are_indexed=rays_are_indexed, ray_angle_res=ray_angle_res,
        rotation=rotation, tilt=tilt, roll=roll, drift=drift, heading=heading,
        pitch=pitch, georefs_applied=georefs_applied)


def _find_all_meta_group_vars(ncvars, meta_group_name):
    """
    Return a list of all variables which are in a given meta_group.
    """
    return [k for k, v in ncvars.iteritems() if 'meta_group' in v.ncattrs()
            and v.meta_group == meta_group_name]


def _ncvar_to_dict(ncvar):
    """ Convert a NetCDF Dataset variable to a dictionary. """
    # copy all attribute except for scaling parameters
    d = dict((k, getattr(ncvar, k)) for k in ncvar.ncattrs()
             if k not in ['scale_factor', 'add_offset'])
    d['data'] = ncvar[:]
    if np.isscalar(d['data']):
        # netCDF4 1.1.0+ returns a scalar for 0-dim array, we always want
        # 1-dim+ arrays with a valid shape.
        d['data'] = np.array(d['data'])
        d['data'].shape = (1, )
    return d


def _unpack_variable_gate_field_dic(
        dic, shape, ray_n_gates, ray_start_index):
    """ Create a 2D array from a 1D field data, dic update in place """
    fdata = dic['data']
    data = np.ma.masked_all(shape, dtype=fdata.dtype)
    for i, (gates, idx) in enumerate(zip(ray_n_gates, ray_start_index)):
        data[i, :gates] = fdata[idx:idx+gates]
    dic['data'] = data
    return


def write_cfradial(filename, radar, format='NETCDF4', time_reference=None,
                   arm_time_variables=False):
    """
    Write a Radar object to a CF/Radial compliant netCDF file.

    The files produced by this routine follow the `CF/Radial standard`_.
    Attempts are also made to to meet many of the standards outlined in the
    `ARM Data File Standards`_.

    .. _CF/Radial standard: http://www.ral.ucar.edu/projects/titan/docs/radial_formats/cfradial.html
    .. _ARM Data File Standards: https://docs.google.com/document/d/1gBMw4Kje6v8LBlsrjaGFfSLoU0jRx-07TIazpthZGt0/edit?pli=1

    Parameters
    ----------
    filename : str
        Filename to create.
    radar : Radar
        Radar object.
    format : str, optional
        NetCDF format, one of 'NETCDF4', 'NETCDF4_CLASSIC',
        'NETCDF3_CLASSIC' or 'NETCDF3_64BIT'. See netCDF4 documentation for
        details.
    time_reference : bool
        True to include a time_reference variable, False will not include
        this variable. The default, None, will include the time_reference
        variable when the first time value is non-zero.
    arm_time_variables : bool
        True to create the ARM standard time variables base_time and
        time_offset, False will not create these variables.

    """
    dataset = netCDF4.Dataset(filename, 'w', format=format)

    # determine the maximum string length
    max_str_len = len(radar.sweep_mode['data'][0])
    for k in ['follow_mode', 'prt_mode', 'polarization_mode']:
        if ((radar.instrument_parameters is not None) and
                (k in radar.instrument_parameters)):
            sdim_length = len(radar.instrument_parameters[k]['data'][0])
            max_str_len = max(max_str_len, sdim_length)
    str_len = max(max_str_len, 32)      # minimum string legth of 32

    # create time, range and sweep dimensions
    dataset.createDimension('time', None)
    dataset.createDimension('range', radar.ngates)
    dataset.createDimension('sweep', radar.nsweeps)
    dataset.createDimension('string_length', str_len)

    # global attributes
    # remove global variables from copy of metadata
    metadata_copy = dict(radar.metadata)
    global_variables = ['volume_number', 'platform_type', 'instrument_type',
                        'primary_axis', 'time_coverage_start',
                        'time_coverage_end', 'time_reference']
    for var in global_variables:
        if var in metadata_copy:
            metadata_copy.pop(var)

    # determine the history attribute if it doesn't exist, save for
    # the last attribute.
    if 'history' in metadata_copy:
        history = metadata_copy.pop('history')
    else:
        user = getpass.getuser()
        node = platform.node()
        time_str = datetime.datetime.now().isoformat()
        t = (user, node, time_str)
        history = 'created by %s on %s at %s using Py-ART' % (t)

    dataset.setncatts(metadata_copy)

    if 'Conventions' not in dataset.ncattrs():
        dataset.setncattr('Conventions', "CF/Radial")

    if 'field_names' not in dataset.ncattrs():
        dataset.setncattr('field_names', ', '.join(radar.fields.keys()))

    # history should be the last attribute, ARM standard
    dataset.setncattr('history', history)

    # arm time variables base_time and time_offset if requested
    if arm_time_variables:
        dt = netCDF4.num2date(radar.time['data'][0], radar.time['units'])
        td = dt - datetime.datetime.utcfromtimestamp(0)
        base_time = {
            'data': np.array([td.seconds + td.days * 24 * 3600], 'int32'),
            'string': dt.strftime('%d-%b-%Y,%H:%M:%S GMT'),
            'units': 'seconds since 1970-1-1 0:00:00 0:00',
            'ancillary_variables': 'time_offset',
            'long_name': 'Base time in Epoch',
        }
        _create_ncvar(base_time, dataset, 'base_time', ())

        time_offset = {
            'data': radar.time['data'],
            'long_name': 'Time offset from base_time',
            'units': radar.time['units'].replace('T', ' ').replace('Z', ''),
            'ancillary_variables': 'time_offset',
            'calendar': 'gregorian',
        }
        _create_ncvar(time_offset, dataset, 'time_offset', ('time', ))

    # standard variables
    _create_ncvar(radar.time, dataset, 'time', ('time', ))
    _create_ncvar(radar.range, dataset, 'range', ('range', ))
    _create_ncvar(radar.azimuth, dataset, 'azimuth', ('time', ))
    _create_ncvar(radar.elevation, dataset, 'elevation', ('time', ))

    # optional sensor pointing variables
    if radar.scan_rate is not None:
        _create_ncvar(radar.scan_rate, dataset, 'scan_rate', ('time', ))
    if radar.antenna_transition is not None:
        _create_ncvar(radar.antenna_transition, dataset,
                      'antenna_transition', ('time', ))

    # fields
    for field, dic in radar.fields.iteritems():
        _create_ncvar(dic, dataset, field, ('time', 'range'))

    # sweep parameters
    _create_ncvar(radar.sweep_number, dataset, 'sweep_number', ('sweep', ))
    _create_ncvar(radar.fixed_angle, dataset, 'fixed_angle', ('sweep', ))
    _create_ncvar(radar.sweep_start_ray_index, dataset,
                  'sweep_start_ray_index', ('sweep', ))
    _create_ncvar(radar.sweep_end_ray_index, dataset,
                  'sweep_end_ray_index', ('sweep', ))
    _create_ncvar(radar.sweep_mode, dataset, 'sweep_mode',
                  ('sweep', 'string_length'))
    if radar.target_scan_rate is not None:
        _create_ncvar(radar.target_scan_rate, dataset, 'target_scan_rate',
                      ('sweep', ))
    if radar.rays_are_indexed is not None:
        _create_ncvar(radar.rays_are_indexed, dataset, 'rays_are_indexed',
                      ('sweep', 'string_length'))
    if radar.ray_angle_res is not None:
        _create_ncvar(radar.ray_angle_res, dataset, 'ray_angle_res',
                      ('sweep', ))

    # instrument_parameters
    if ((radar.instrument_parameters is not None) and
            ('frequency' in radar.instrument_parameters.keys())):
        size = len(radar.instrument_parameters['frequency']['data'])
        dataset.createDimension('frequency', size)

    if radar.instrument_parameters is not None:
        for k in radar.instrument_parameters.keys():
            if k in _INSTRUMENT_PARAMS_DIMS:
                dim = _INSTRUMENT_PARAMS_DIMS[k]
            else:
                dim = ()
            _create_ncvar(radar.instrument_parameters[k], dataset, k, dim)

    # radar_calibration variables
    if radar.radar_calibration is not None and radar.radar_calibration != {}:
        size = [len(d['data']) for k, d in radar.radar_calibration.items()
                if k not in ['r_calib_index', 'r_calib_time']][0]
        dataset.createDimension('r_calib', size)
        for key, dic in radar.radar_calibration.items():
            if key == 'r_calib_index':
                dims = ('time', )
            elif key == 'r_calib_time':
                dims = ('r_calib', 'string_length')
            else:
                dims = ('r_calib', )
            _create_ncvar(dic, dataset, key, dims)

    # latitude, longitude, altitude, altitude_agl
    if radar.latitude['data'].size == 1:
        # stationary platform
        _create_ncvar(radar.latitude, dataset, 'latitude', ())
        _create_ncvar(radar.longitude, dataset, 'longitude', ())
        _create_ncvar(radar.altitude, dataset, 'altitude', ())
        if radar.altitude_agl is not None:
            _create_ncvar(radar.altitude_agl, dataset, 'altitude_agl', ())
    else:
        # moving platform
        _create_ncvar(radar.latitude, dataset, 'latitude', ('time', ))
        _create_ncvar(radar.longitude, dataset, 'longitude', ('time', ))
        _create_ncvar(radar.altitude, dataset, 'altitude', ('time', ))
        if radar.altitude_agl is not None:
            _create_ncvar(radar.altitude_agl, dataset, 'altitude_agl',
                          ('time', ))

    # time_coverage_start and time_coverage_end variables
    time_dim = ('string_length', )
    units = radar.time['units']
    start_dt = netCDF4.num2date(radar.time['data'][0], units)
    end_dt = netCDF4.num2date(radar.time['data'][-1], units)
    start_dic = {'data': np.array(start_dt.isoformat() + 'Z'),
                 'long_name': 'UTC time of first ray in the file',
                 'units': 'unitless'}
    end_dic = {'data': np.array(end_dt.isoformat() + 'Z'),
               'long_name': 'UTC time of last ray in the file',
               'units': 'unitless'}
    _create_ncvar(start_dic, dataset, 'time_coverage_start', time_dim)
    _create_ncvar(end_dic, dataset, 'time_coverage_end', time_dim)

    # time_reference is required or requested.
    if time_reference is None:
        if radar.time['data'][0] == 0:
            time_reference = False
        else:
            time_reference = True
    if time_reference:
        ref_dic = {'data': np.array(radar.time['units'][-20:], dtype='S'),
                   'long_name': 'UTC time reference',
                   'units': 'unitless'}
        _create_ncvar(ref_dic, dataset, 'time_reference', time_dim)

    # global variables
    # volume_number, required
    vol_dic = {'long_name': 'Volume number', 'units': 'unitless'}
    if 'volume_number' in radar.metadata:
        vol_dic['data'] = np.array([radar.metadata['volume_number']],
                                   dtype='int32')
    else:
        vol_dic['data'] = np.array([0], dtype='int32')
    _create_ncvar(vol_dic, dataset, 'volume_number', ())

    # platform_type, optional
    if 'platform_type' in radar.metadata:
        dic = {'long_name': 'Platform type',
               'data': np.array(radar.metadata['platform_type'])}
        _create_ncvar(dic, dataset, 'platform_type', ('string_length', ))

    # instrument_type, optional
    if 'instrument_type' in radar.metadata:
        dic = {'long_name': 'Instrument type',
               'data': np.array(radar.metadata['instrument_type'])}
        _create_ncvar(dic, dataset, 'instrument_type', ('string_length', ))

    # primary_axis, optional
    if 'primary_axis' in radar.metadata:
        dic = {'long_name': 'Primary axis',
               'data': np.array(radar.metadata['primary_axis'])}
        _create_ncvar(dic, dataset, 'primary_axis', ('string_length', ))

    # moving platform geo-reference variables
    if radar.rotation is not None:
        _create_ncvar(radar.rotation, dataset, 'rotation', ('time', ))

    if radar.tilt is not None:
        _create_ncvar(radar.tilt, dataset, 'tilt', ('time', ))

    if radar.roll is not None:
        _create_ncvar(radar.roll, dataset, 'roll', ('time', ))

    if radar.drift is not None:
        _create_ncvar(radar.drift, dataset, 'drift', ('time', ))

    if radar.heading is not None:
        _create_ncvar(radar.heading, dataset, 'heading', ('time', ))

    if radar.pitch is not None:
        _create_ncvar(radar.pitch, dataset, 'pitch', ('time', ))

    if radar.georefs_applied is not None:
        _create_ncvar(radar.georefs_applied, dataset, 'georefs_applied',
                      ('time', ))

    dataset.close()


def _create_ncvar(dic, dataset, name, dimensions):
    """
    Create and fill a Variable in a netCDF Dataset object.

    Parameters
    ----------
    dic : dict
        Radar dictionary to containing variable data and meta-data
    dataset : Dataset
        NetCDF dataset to create variable in.
    name : str
        Name of variable to create.
    dimension : tuple of str
        Dimension of variable.

    """
    # create array from list, etc.
    data = dic['data']
    if isinstance(data, np.ndarray) is not True:
        print "Warning, converting non-array to array:", name
        data = np.array(data)

    # convert string array to character arrays
    if data.dtype.char is 'S' and data.dtype != 'S1':
        data = stringarray_to_chararray(data)
    if data.dtype.char is 'U' and data.dtype != 'U1':
        data = stringarray_to_chararray(data)

    # create the dataset variable
    if 'least_significant_digit' in dic:
        lsd = dic['least_significant_digit']
    else:
        lsd = None
    if "_FillValue" in dic:
        fill_value = dic['_FillValue']
    else:
        fill_value = None

    ncvar = dataset.createVariable(name, data.dtype, dimensions,
                                   zlib=True, least_significant_digit=lsd,
                                   fill_value=fill_value)

    # long_name attribute first if present, ARM standard
    if 'long_name' in dic.keys():
        ncvar.setncattr('long_name', dic['long_name'])

    # units attribute second if present, ARM standard
    if 'units' in dic.keys():
        ncvar.setncattr('units', dic['units'])

    # remove _FillValue and replace to make it the third attribute.
    if '_FillValue' in ncvar.ncattrs():
        fv = ncvar._FillValue
        ncvar.delncattr('_FillValue')
        ncvar.setncattr('_FillValue', fv)

    # set all attributes
    for key, value in dic.iteritems():
        if key not in ['data', '_FillValue', 'long_name', 'units']:
            ncvar.setncattr(key, value)

    # set the data
    if data.shape == ():
        data.shape = (1,)
    if data.dtype == 'S1':  # string/char arrays
        ncvar[..., :data.shape[-1]] = data[:]
    else:
        ncvar[:] = data[:]



"""
pyart.correct.dealias
=====================

Front end to the University of Washington 4DD code for Doppler dealiasing.

.. autosummary::
    :toctree: generated/

    dealias_fourdd
    find_time_in_interp_sonde
    _create_rsl_volume

"""
# Nothing from this module is imported to pyart.correct if RSL is not
# installed.

import numpy as np

from ..config import get_field_name, get_fillvalue, get_metadata
from ..io import _rsl_interface
from . import _fourdd_interface
from ..util import datetime_utils


def dealias_fourdd(radar, last_radar=None, sounding_heights=None,
                   sounding_wind_speeds=None, sounding_wind_direction=None,
                   prep=1, filt=1, rsl_badval=131072.0, keep_original=False,
                   extra_prep=False, ncp_min=0.3, rhv_min=0.7,
                   refl_field=None, vel_field=None, corr_vel_field=None,
                   last_vel_field=None, ncp_field=None, rhv_field=None,
                   debug=False, max_shear=0.05, sign=1, **kwargs):
    """
    Dealias Doppler velocities using the 4DD algorithm.

    Dealias the Doppler velocities field using the University of Washington
    4DD algorithm utilizing information from a previous volume scan and/or
    sounding data. Either last_radar or sounding_heights,
    sounding_wind_speeds and sounding_wind_direction must be provided.
    For best results provide both a previous volume scan and sounding data.
    Radar and last_radar must contain the same number of rays per sweep.

    Additional arguments are passed to
    :py:func:`_fourdd_interface.fourdd_dealias`.
    These can be used to fine tune the behavior of the FourDD algorithm.
    See the documentation of Other Parameters for details.

    Parameters
    ----------
    radar : Radar
        Radar object to use for dealiasing.  Must have a Nyquist defined in
        the instrument_parameters attribute and have a
        reflectivity_horizontal and mean_doppler_velocity fields.
    last_radar : Radar, optional
        The previous radar volume, which has been successfully
        dealiased. Using a previous volume as an initial condition can
        greatly improve the dealiasing, and represents the final dimension
        in the 4DD algorithm.
    sounding_heights : ndarray, optional
        Sounding heights is meters above mean sea level.  If altitude
        attribute of the radar object if reference against something other
        than mean sea level then this parameter should also be referenced in
        that manner.
    sounding_wind_speeds : ndarray, optional
        Sounding wind speeds in m/s.
    sounding_wind_direction : ndarray, optional
        Sounding wind directions in degrees.

    Other Parameters
    ----------------
    prep : int
        Flag controlling thresholding, 1 = yes, 0 = no.
    filt : int
        Flag controlling Bergen and Albers filter, 1 = yes, 0 = no.
    rsl_badval : float
        Value which represents a bad value in RSL.
    keep_original : bool
        True to keep original doppler velocity values when the dealiasing
        procedure fails, otherwises these gates will be masked.  NaN values
        are still masked.
    extra_prep : bool
        True to use extra volume preperation, which will use the normalized
        coherent power and cross correlation ratio fields to further remove
        bad gates. Set this to False (default) if those fields are not
        available.
    ncp_min : float
        Minimum normalized coherent power allowed for any gate. Only
        applicable when extra_prep is True.
    rhv_min : float
        Minimum cross correlation ratio allowed for any gate. Only
        applicable when extra_prep is True.
    refl_field : str
        Field in radar to use as the reflectivity during dealiasing.
        None will use the default field name from the Py-ART configuration
        file.
    vel_field : str
        Field in radar to use as the Doppler velocities during dealiasing.
        None will use the default field name from the Py-ART configuration
        file.
    corr_vel_field : str
        Name to use for the dealiased Doppler velocity field metadata.  None
        will use the default field name from the Py-ART configuration file.
    last_vel_field : str
        Name to use for the dealiased Doppler velocity field metadata in
        last_radar.  None will use the corr_vel_field name.
    ncp_field, rhv_field : str
        Fields in radar to use as the normalized coherent power and cross
        correlation ratio, respectively. None will use the default field name
        from the Py-ART configuration file. Only applicable when extra_prep
        is True.
    maxshear : float
        Maximum vertical shear which will be incorperated into the created
        volume from the sounding data.  Parameter not used when no
        sounding data is provided.
    sign : int
        Sign convention which the radial velocities in the volume created
        from the sounding data will will. This should match the convention
        used in the radar data. A value of 1 represents when positive values
        velocities are towards the radar, -1 represents when negative
        velocities are towards the radar.
    debug : bool
        Set True to return RSL Volume objects for debugging:
        usuccess, DZvolume, radialVelVolume, unfoldedVolume, sondVolume

    Returns
    -------
    vr_corr : dict
        Field dictionary containing dealiased Doppler velocities.  Dealiased
        array is stored under the 'data' key.

    Notes
    -----
    Due to limitations in the C code do not call with sounding arrays over
    999 elements long.

    References
    ----------
    C. N. James and R. A Houze Jr, A Real-Time Four-Dimensional Doppler
    Dealising Scheme, Journal of Atmospheric and Oceanic Technology, 2001, 18,
    1674.

    """
    # TODO test with RHI radar scan

    # verify that sounding data or last_volume is provided
    sounding_available = ((sounding_heights is not None) and
                          (sounding_wind_speeds is not None) and
                          (sounding_wind_direction is not None))
    if (not sounding_available) and (last_radar is None):
        raise ValueError('sounding data or last_radar must be provided')

    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if vel_field is None:
        vel_field = get_field_name('velocity')
    if corr_vel_field is None:
        corr_vel_field = get_field_name('corrected_velocity')
    if last_vel_field is None:
        last_vel_field = get_field_name('corrected_velocity')
    if ncp_field is None:
        ncp_field = get_field_name('normalized_coherent_power')
    if rhv_field is None:
        rhv_field = get_field_name('cross_correlation_ratio')

    # get fill value
    fill_value = get_fillvalue()

    # check extra volume preparation inputs
    if extra_prep:
        if ncp_field not in radar.fields and rhv_field not in radar.fields:
            raise KeyError(('Radar does not have necessary fields for'
                            'extra volume preparation'))

    # extra volume preparation
    # this assumes the radar has a normalized coherent power field and a
    # cross correlation ratio field
    if prep and extra_prep:
        not_coherent = np.logical_or(
            radar.fields[ncp_field]['data'] < ncp_min,
            radar.fields[rhv_field]['data'] < rhv_min)
        fdata = np.copy(radar.fields[refl_field]['data']).astype(np.float32)
        fdata[not_coherent] = fill_value
    else:
        fdata = None

    # create RSL volumes containing the reflectivity, doppler velocity, and
    # doppler velocity in the last radar (if provided)
    refl_volume = _create_rsl_volume(radar, refl_field, 0, rsl_badval, fdata)
    vel_volume = _create_rsl_volume(radar, vel_field, 1, rsl_badval)

    if last_radar is not None:
        last_vel_volume = _create_rsl_volume(
            last_radar, last_vel_field, 1, rsl_badval)
    else:
        last_vel_volume = None

    # create an RslVolume containing the sounding data if it available
    if sounding_available:
        # convert the sounding data to 1D float32 arrays
        hc = np.ascontiguousarray(sounding_heights, dtype=np.float32)
        sc = np.ascontiguousarray(sounding_wind_speeds, dtype=np.float32)
        dc = np.ascontiguousarray(sounding_wind_direction, dtype=np.float32)

        success, sound_volume = _fourdd_interface.create_soundvolume(
            refl_volume, hc, sc, dc, sign, max_shear)
        if success == 0:
            raise ValueError('Error when loading sounding data')
    else:
        sound_volume = None

    # perform dealiasing
    if debug:
        return _fourdd_interface.fourdd_dealias(
            vel_volume, last_vel_volume, sound_volume, refl_volume,
            prep, filt, debug=True, **kwargs)

    flag, data = _fourdd_interface.fourdd_dealias(
        vel_volume, last_vel_volume, sound_volume, refl_volume, prep, filt,
        debug=False, **kwargs)

    # prepare data for output, set bad values and mask data
    is_bad_data = np.logical_or(np.isnan(data), data == rsl_badval)
    if keep_original:
        vel_array = radar.fields[vel_field]['data']
        data = np.where(is_bad_data, vel_array, data)
    else:
        data[is_bad_data] = fill_value
    data = np.ma.masked_equal(data, fill_value)

    # return field dictionary containing dealiased Doppler velocities
    vr_corr = get_metadata(corr_vel_field)
    vr_corr['data'] = data
    vr_corr['_FillValue'] = data.fill_value
    return vr_corr


def _create_rsl_volume(radar, field_name, vol_num, rsl_badval, fdata=None):
    """
    Create a RSLVolume containing data from a field in radar.
    """
    fill_value = get_fillvalue()
    if fdata is None:
        fdata = np.copy(radar.fields[field_name]['data']).astype(np.float32)
    fdata = np.ma.filled(fdata, fill_value)
    is_bad = np.logical_or(fdata == fill_value, np.isnan(fdata))
    fdata[is_bad] = rsl_badval
    rays_per_sweep = (radar.sweep_end_ray_index['data'] -
                      radar.sweep_start_ray_index['data'] + 1)
    rays_per_sweep = rays_per_sweep.astype(np.int32)
    rsl_volume = _rsl_interface.create_volume(fdata, rays_per_sweep, vol_num)
    _rsl_interface._label_volume(rsl_volume, radar)
    return rsl_volume


def find_time_in_interp_sonde(interp_sonde, target, debug=False):
    """
    Find the wind parameter for a given time in a ARM interpsonde file.

    Parameters
    ----------
    interp_sonde : netCDF4.Dataset
        netCDF4 object pointing to a ARM interpsonde file.
    target : datetime
        Target datetime, the closest time in the interpsonde file will be
        used.

    Other Parameters
    ----------------
    debug : bool
        Print debugging information.

    Returns
    -------
    height : np.ndarray
        Heights above the ground for the time closest to target.
    speed : np.ndarray
        Wind speeds at given height for the time closest to taget.
    direction : np.ndarray
        Wind direction at given height for the time closest to target.

    """

    sonde_datetimes = datetime_utils.datetimes_from_dataset(interp_sonde)

    # get closest time index to target
    idx = np.abs(sonde_datetimes - target).argmin()

    if debug:
        print 'Target time is %s' % (target)
        print 'Interpolated sounding time is %s' % (sonde_datetimes[idx])

    return (interp_sonde.variables['height'][:],
            interp_sonde.variables['wspd'][idx, :],
            interp_sonde.variables['wdir'][idx, :])



"""
pyart.graph.plot_cfradial
=========================

Routines for plotting radar data from CF/Radial netCDF files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    CFRadialDisplay

"""

import numpy as np
import netCDF4

from .radardisplay import RadarDisplay
from .common import radar_coords_to_cart


class CFRadialDisplay(RadarDisplay):
    """
    A display object for creating plots from data in NetCDF4 Dataset objects.

    Parameters
    ----------
    dataset : Dataset
        NetCDF4 Dataset object to use for creating plots.
    shift : (float, float)
        Shifts in km to offset the calulcated x and y locations.

    Attributes
    ----------
    plots : list
        List of plots created.
    plot_vars : list
        List of fields plotted, order matches plot list.
    cbs : list
        List of colorbars created.
    radar_name : str
        Name of radar.
    origin : str
        'Origin' or 'Radar'.
    shift : (float, float)
        Shift in meters.
    x, y, z : array
        Cartesian location of a sweep in meters.
    loc : (float, float)
        Latitude and Longitude of radar in degrees.
    time_begin : datetime
        Beginning time of first radar scan.
    starts : array
        Starting ray index for each sweep.
    ends : array
        Ending ray index for each sweep.
    ranges : array
        Gate ranges in meters.
    azimuths : array
        Azimuth angle in degrees.
    elevations : array
        Elevations in degrees.
    fixed_angle : array
        Scan angle in degrees.

    """
    # missing scan_type and fields attributes
    def __init__(self, dataset, shift=(0.0, 0.0)):
        """ Initialize the object. """

        # populate attributes from dataset
        self.dataset = dataset
        self.ranges = dataset.variables['range']
        self.azimuths = dataset.variables['azimuth']
        self.elevations = dataset.variables['elevation']
        self.fixed_angle = dataset.variables['fixed_angle']
        self.antenna_transition = None
        if 'instrument_name' in dataset.ncattrs():
            self.radar_name = dataset.instrument_name
        else:
            self.radar_name = ''

        # origin
        if shift != (0.0, 0.0):
            self.origin = 'origin'
        else:
            self.origin = 'radar'

        # x, y, z attributes: cartesian location for a sweep in km.
        rg, azg = np.meshgrid(self.ranges, self.azimuths)
        rg, eleg = np.meshgrid(self.ranges, self.elevations)
        self.shift = shift
        self.x, self.y, self.z = radar_coords_to_cart(rg / 1000.0, azg, eleg)
        self.x = self.x + self.shift[0]
        self.y = self.y + self.shift[1]

        # radar location in latitude and longitude
        lon = dataset.variables['longitude'][:]
        lat = dataset.variables['latitude'][:]
        self.loc = [lat, lon]

        # datetime object describing first sweep time
        t = dataset.variables['time']
        self.time_begin = netCDF4.num2date(t[0], t.units, t.calendar)

        # sweep start and end indices
        self.starts = dataset.variables['sweep_start_ray_index'][:]
        self.ends = dataset.variables['sweep_end_ray_index'][:]

        # list to hold plots, plotted fields and plotted colorbars
        self.plots = []
        self.plot_vars = []
        self.cbs = []

    # public methods which are overridden.

    def generate_title(self, field, sweep):
        """
        Generate a title for a plot.

        Parameters
        ----------
        field : str
            Field plotted.
        sweep : int
            Sweep plotted.

        Returns
        -------
        title : str
            Plot title.

        """
        time_str = self.time_begin.isoformat() + 'Z'
        fixed_angle = self.fixed_angle[sweep]
        l1 = "%s %.1f Deg. %s " % (self.radar_name, fixed_angle, time_str)

        field_name = self.dataset.variables[field].standard_name
        field_name = field_name.replace('_', ' ')
        field_name = field_name[0].upper() + field_name[1:]
        return l1 + '\n' + field_name

    # private methods which are overridden

    def _parse_vmin_vmax(self, field, vmin, vmax):
        if vmin is None:
            vmin = self.dataset.variables[field].valid_min
        if vmax is None:
            vmax = self.dataset.variables[field].valid_max
        return vmin, vmax

    def _get_data(self, field, sweep, mask_tuple, filter_transitions):
        """ Retrieve and return data from a plot function. """
        start = self.starts[sweep]
        end = self.ends[sweep] + 1
        data = self.dataset.variables[field][start:end]

        if mask_tuple is not None:
            mask_field, mask_value = mask_tuple
            mdata = self.dataset.variables[mask_field][start:end]
            data = np.ma.masked_where(mdata < mask_value, data)
        return data

    def _get_colorbar_label(self, field):
        """ Return a colobar label for a given field. """
        last_field_dict = self.dataset.variables[field].ncattrs()
        if 'standard_name' in last_field_dict:
            standard_name = self.dataset.variables[field].standard_name
        else:
            standard_name = self.dataset.variables[field].long_name
        units = self.dataset.variables[field].units
        return self._generate_colorbar_label(standard_name, units)
